//-------------------------------------------------------------------------------------------------
// LineSegment2.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_LineSegment2_
#define _include_LineSegment2_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "AABB2.hpp"
#include "engineCommon.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		LineSegment2
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class LineSegment2
{
public:
	Vector2		start;
	Vector2		displacement;

	/// Construction / Destruction
	~LineSegment2(); // destructor
	LineSegment2(); // default constructor
	LineSegment2( const LineSegment2& LineSegment2 ); // copy constructor
	LineSegment2( const Vector2& point );
	LineSegment2( const Vector2& start, const Vector2& displacement );
	LineSegment2( float startX, float startY, float displacementX, float displacementY );

	/// Interface methods
	void SetFromStartDisplacement( const Vector2& start, const Vector2& displacement );
	void SetFromStartXYDisplacementXY( float startX, float startY, float displacementX, float displacementY );
	void SetFromStartEnd( const Vector2& start, const Vector2& end );
	void SetFromStartXYEndXY( float startX, float startY, float endX, float endY );
	void PushNewDisplacement( const Vector2& newDisplacement );
	const LineSegment2 Lerp( const LineSegment2& lerpTo, const float lerpTo_fraction ) const;
	const float CalcLengthSquared() const;
	const float CalcLength() const;
	const Vector2& GetDimensions() const;
	const AABB2 CalcExtents() const;

	/// Operators
	const LineSegment2& operator = ( const LineSegment2& rhs );
	const LineSegment2& operator += ( const Vector2& translation );
	const LineSegment2& operator -= ( const Vector2& translation );
	const LineSegment2& operator *= ( const float scale );
	const LineSegment2& operator /= ( const float divisor );

	const LineSegment2 operator - () const;
	const LineSegment2 operator * ( const float scale ) const;

	const bool operator == ( const LineSegment2& rhs ) const;
	const bool operator != ( const LineSegment2& rhs ) const;
};

//-------------------------------------------------------------------------------------------------
// External Operators & Functions
//-------------------------------------------------------------------------------------------------
const LineSegment2 operator * ( const float scale, const LineSegment2& rhs );
const bool EpsilonEqual( const LineSegment2& lhs, const LineSegment2& rhs, const float epsilon );


//-------------------------------------------------------------------------------------------------
// ~LineSegment2() - destructor
//-------------------------------------------------------------------------------------------------
inline LineSegment2::~LineSegment2()
{
}


//-------------------------------------------------------------------------------------------------
// LineSegment2() - default constructor
//-------------------------------------------------------------------------------------------------
inline LineSegment2::LineSegment2()
{
	// do nothing - keep this class as flyweight as possible
}


//-------------------------------------------------------------------------------------------------
// LineSegment2( const LineSegment2& LineSegment2 ) - copy constructor
//-------------------------------------------------------------------------------------------------
inline LineSegment2::LineSegment2( const LineSegment2& LineSegment2 )
:
start( LineSegment2.start ),
displacement( LineSegment2.displacement )
{
}


//-------------------------------------------------------------------------------------------------
// LineSegment2( Vector2 )
//-------------------------------------------------------------------------------------------------
inline LineSegment2::LineSegment2( const Vector2& point )
:
start( point ),
displacement( Vector2::ZERO )
{
}


//-------------------------------------------------------------------------------------------------
// LineSegment2( Vector2, Vector2 )
//-------------------------------------------------------------------------------------------------
inline LineSegment2::LineSegment2( const Vector2& start, const Vector2& displacement )
:
start( start ),
displacement( displacement )
{
}


//-------------------------------------------------------------------------------------------------
// LineSegment2( float, float, float, float )
//-------------------------------------------------------------------------------------------------
inline LineSegment2::LineSegment2( float startX, float startY, float displacementX, float displacementY )
:
start( startX, startY ),
displacement( displacementX, displacementY )
{
}


//-------------------------------------------------------------------------------------------------
// SetFromStartDisplacement
//-------------------------------------------------------------------------------------------------
inline void LineSegment2::SetFromStartDisplacement( const Vector2& startPoint, const Vector2& displacementVector )
{
	start = startPoint;
	displacement = displacementVector;
}


//-------------------------------------------------------------------------------------------------
// SetFromStartEnd
//-------------------------------------------------------------------------------------------------
inline void LineSegment2::SetFromStartEnd( const Vector2& startPoint, const Vector2& endPoint )
{
	start = startPoint;
	displacement = endPoint - startPoint;
}


//-------------------------------------------------------------------------------------------------
// SetFromStartXYDisplacementXY
//-------------------------------------------------------------------------------------------------
inline void LineSegment2::SetFromStartXYDisplacementXY( float startX, float startY, float displacementX, float displacementY )
{
	start.SetXY( startX, startY );
	displacement.SetXY( displacementX, displacementY );
}


//-------------------------------------------------------------------------------------------------
// SetFromStartXYEndXY
//-------------------------------------------------------------------------------------------------
inline void LineSegment2::SetFromStartXYEndXY( float startX, float startY, float endX, float endY )
{
	start.SetXY( startX, startY );
	displacement.SetXY( endX - startX, endY - startY );
}


//-------------------------------------------------------------------------------------------------
// PushNewDisplacement
//-------------------------------------------------------------------------------------------------
inline void LineSegment2::PushNewDisplacement( const Vector2& newDisplacement )
{
	/// Set the new start at the old endpoint
	start += displacement;

	/// Assign a new displacement
	displacement = newDisplacement;
}


//-------------------------------------------------------------------------------------------------
// operator =
//-------------------------------------------------------------------------------------------------
inline const LineSegment2& LineSegment2::operator = ( const LineSegment2& rhs )
{
	if( this == &rhs )
		return *this;

	start = rhs.start;
	displacement = rhs.displacement;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator +=
//-------------------------------------------------------------------------------------------------
inline const LineSegment2& LineSegment2::operator += ( const Vector2& translation )
{
	start += translation;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator -=
//-------------------------------------------------------------------------------------------------
inline const LineSegment2& LineSegment2::operator -= ( const Vector2& translation )
{
	start -= translation;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator *=
//-------------------------------------------------------------------------------------------------
inline const LineSegment2& LineSegment2::operator *= ( const float scale )
{
	displacement *= scale;
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator /=
//-------------------------------------------------------------------------------------------------
inline const LineSegment2& LineSegment2::operator /= ( const float divisor )
{
	if( divisor )
	{
		displacement /= divisor;
	}
	else
	{
		displacement = Vector2::ZERO;
	}
	return *this;
}


//-------------------------------------------------------------------------------------------------
// operator - (unary minus)
//-------------------------------------------------------------------------------------------------
inline const LineSegment2 LineSegment2::operator - () const
{
	return LineSegment2( start, -displacement );
}


//-------------------------------------------------------------------------------------------------
// operator *
//-------------------------------------------------------------------------------------------------
inline const LineSegment2 LineSegment2::operator * ( const float scale ) const
{
	return( LineSegment2( start, displacement * scale ) );
}


//-------------------------------------------------------------------------------------------------
// operator ==
//-------------------------------------------------------------------------------------------------
inline const bool LineSegment2::operator == ( const LineSegment2& rhs ) const
{
	return( (start == rhs.start) && (displacement == rhs.displacement) );
}


//-------------------------------------------------------------------------------------------------
// operator !=
//-------------------------------------------------------------------------------------------------
inline const bool LineSegment2::operator != ( const LineSegment2& rhs ) const
{
	return( (start != rhs.start) || (displacement != rhs.displacement) );
}


//-------------------------------------------------------------------------------------------------
// CalcDimensions
//-------------------------------------------------------------------------------------------------
inline const Vector2& LineSegment2::GetDimensions() const
{
	return displacement;
}


//-------------------------------------------------------------------------------------------------
// CalcExtents
//-------------------------------------------------------------------------------------------------
inline const AABB2 LineSegment2::CalcExtents() const
{
	Vector2 end = start + displacement;
	if( end.x > start.x )
	{
		if( end.y > start.y )
		{
			return AABB2( start.x, start.y, end.x, end.y );
		}
		else
		{
			return AABB2( start.x, end.y, end.x, start.y );
		}
	}
	else
	{
		if( end.y > start.y )
		{
			return AABB2( end.x, start.y, start.x, end.y );
		}
		else
		{
			return AABB2( end.x, end.y, start.x, start.y );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// CalcLengthSquared
//-------------------------------------------------------------------------------------------------
inline const float LineSegment2::CalcLengthSquared() const
{
	return displacement.CalcLengthSquared();
}


//-------------------------------------------------------------------------------------------------
// CalcLength
//-------------------------------------------------------------------------------------------------
inline const float LineSegment2::CalcLength() const
{
	return displacement.CalcLength();
}


//-------------------------------------------------------------------------------------------------
// Lerp
//-------------------------------------------------------------------------------------------------
inline const LineSegment2 LineSegment2::Lerp( const LineSegment2& lerpTo, const float lerpTo_fraction ) const
{
	return LineSegment2( start + lerpTo_fraction * (lerpTo.start - start),
					displacement + lerpTo_fraction * (lerpTo.displacement - displacement));
}


//-------------------------------------------------------------------------------------------------
// operator * (scale, LineSegment2)
//-------------------------------------------------------------------------------------------------
inline const LineSegment2 operator * ( const float scale, const LineSegment2& rhs )
{
	return( LineSegment2( rhs.start, scale * rhs.displacement ) );
}


//-------------------------------------------------------------------------------------------------
// EpsilonEqual
//-------------------------------------------------------------------------------------------------
inline const bool EpsilonEqual( const LineSegment2& lhs, const LineSegment2& rhs, const float epsilon )
{
	return(	( fabs( lhs.start.x - rhs.start.x ) < epsilon )&&
			( fabs( lhs.start.y - rhs.start.y ) < epsilon ) &&
			( fabs( lhs.displacement.x - rhs.displacement.x ) < epsilon ) &&
			( fabs( lhs.displacement.y - rhs.displacement.y ) < epsilon ) );
}


#endif // _include_LineSegment2_
