//-------------------------------------------------------------------------------------------------
// TheRenderer.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheRenderer.hpp"
#include "TheVideoSystem.hpp"
#include "TheConsole.hpp"
#include "TheApplication.hpp"
#include "Vector2.hpp"
#include "engineAssert.hpp"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Subsystem name macro (for assertions, etc)
//-------------------------------------------------------------------------------------------------
#define SUBSYSTEM_NAME "TheRenderer"


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC bool				TheRenderer::_isInitialized			= false;


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheRenderer::TheRenderer()
:
_zBufferDepthMin( 0.0f ),
_zBufferDepthMax( 1.0f ),
_consoleDisplayMode( CONSOLE_DISPLAY_NONE ),
_displayFPS( false ),
_numConsolePartialRows( 3 ),
_currentTexNum( 0 )
{
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheRenderer::~TheRenderer()
{
	/// Make sure the subsystem was shut down before being destroyed
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was destroyed while running; Shutdown() should be called first!" );
}


//-------------------------------------------------------------------------------------------------
// Startup
//-------------------------------------------------------------------------------------------------
void TheRenderer::Startup()
{
	theConsole.Printf( "::: Initialization : Renderer (starting)\n" );

	/// Make sure the subsystem doesn't get initialized twice
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was initialized twice; either a second "SUBSYSTEM_NAME" was started or Startup() was called on an already-running system." );

	/// Set preferred opengl default states
	_InitOpenGLStates();

	/// Create default font(s) and corresponding display lists
	_InitFonts();

	/// Set initialization complete
	_isInitialized = true;
	theConsole.Printf( "::: Initialization : Renderer (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheRenderer::Shutdown()
{
	theConsole.Printf( "::: Shutdown : Renderer (starting)\n" );

	/// De-initialize the system
	_isInitialized = false;
	theConsole.Printf( "::: Shutdown : Renderer (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// _InitOpenGLStates
//-------------------------------------------------------------------------------------------------
void TheRenderer::_InitOpenGLStates() const
{
	/// Set up default rendering modes
	glShadeModel( GL_FLAT );								// don't use smooth (gouraud) shading
	glClearColor( 0.0f, 0.0f, 0.0f, 0.0f );					// set background/clear color to black
	glClearDepth( _zBufferDepthMax );						// clear z-buffer (set all z-values to 1.0 = max z)
	glDepthFunc( GL_LEQUAL );								// use "less than or equal to" rule for z-test success
	glBlendFunc( GL_SRC_ALPHA, GL_ONE );					// set default alpha blending function
	glLineWidth( 1.0f );									// pixel-width of lines
	glPointSize( 1.0f );									// pixel-width of points
	glFrontFace( GL_CCW );									// clockwise / counter-clockwise winding for front faces
	glCullFace( GL_BACK );									// cull front and/or back polygon faces
	glDisable( GL_ALPHA_TEST );								// alpha testing (reject if alpha below a value, etc)
	glEnable( GL_BLEND );									// blend incoming pixels with those in the framebuffer
	glDisable( GL_CULL_FACE );								// enable front and/or back polygon face culling
	glDisable( GL_DEPTH_TEST );								// perform z-buffer testing before writing a pixel
	glDisable( GL_DITHER );									// dither pixel colors before writing to buffer
	glDisable( GL_FOG );									// perform z-distance-based fog blending
	glDisable( GL_LIGHTING );								// use opengl lighting
	glDisable( GL_LINE_SMOOTH );							// perform pixel antialiasing during line drawing
	glDisable( GL_LINE_STIPPLE );							// use the current line stipple pattern to omit pixels
	glDisable( GL_POINT_SMOOTH );							// perform pixel antialiasing during point drawing
	glDisable( GL_POLYGON_OFFSET_FILL );					// use a z-offset for depth checking (for filled polys)
	glDisable( GL_POLYGON_OFFSET_LINE );					// use a z-offset for depth checking (for outlined polys)
	glDisable( GL_POLYGON_OFFSET_POINT );					// use a z-offset for depth checking (for point-drawn polys)
	glDisable( GL_POLYGON_SMOOTH );							// perform pixel antialiasing during polygon drawing
	glDisable( GL_POLYGON_STIPPLE );						// use the current polygon stipple pattern to omit pixels
	glDisable( GL_SCISSOR_TEST );							// discard fragments that are outside the scissor rectangle
	glDisable( GL_STENCIL_TEST );							// do stencil testing and update the stencil buffer
	glDisable( GL_TEXTURE_2D );								// if enabled, two-dimensional texturing is performed
	glHint( GL_FOG_HINT, GL_FASTEST );						// determines the quality of fog calculations
	glHint( GL_LINE_SMOOTH_HINT, GL_FASTEST );				// determines the quality of line antialiasing
	glHint( GL_PERSPECTIVE_CORRECTION_HINT, GL_FASTEST );	// determines the quality of perspective correction
	glHint( GL_POINT_SMOOTH_HINT, GL_FASTEST );				// determines the quality of point antialiasing
	glHint( GL_POLYGON_SMOOTH_HINT, GL_FASTEST );			// determines the quality of polygon antialiasing
}


//-------------------------------------------------------------------------------------------------
// _InitFonts
//-------------------------------------------------------------------------------------------------
void TheRenderer::_InitFonts()
{
	/// Define the attributes of a true-type font
	LOGFONT fontAttributes;
	memset( &fontAttributes, 0, sizeof( fontAttributes ) );
	fontAttributes.lfHeight = DEFAULT_FONT_HEIGHT;
	fontAttributes.lfWidth = 0;
	fontAttributes.lfEscapement = 0;
	fontAttributes.lfOrientation = 0;
	fontAttributes.lfWeight = FW_NORMAL;
	fontAttributes.lfItalic = FALSE;
	fontAttributes.lfUnderline = FALSE;
	fontAttributes.lfStrikeOut = FALSE;
	fontAttributes.lfCharSet = ANSI_CHARSET;
	fontAttributes.lfOutPrecision = OUT_TT_ONLY_PRECIS;
	fontAttributes.lfClipPrecision = CLIP_DEFAULT_PRECIS;
	fontAttributes.lfQuality = ANTIALIASED_QUALITY;
	fontAttributes.lfPitchAndFamily = FF_SWISS;
	strcpy( fontAttributes.lfFaceName, "Trebuchet MS" );

	/// Create a temporary windows GDI font object with these properties
	HFONT gdiFont = CreateFontIndirect( &fontAttributes );

	/// Get the display context from the video system
	HDC hDC = theVideoSystem.GetDisplayContext();

	/// Select this as the "current" font temporarily
	HFONT oldGdiFont = (HFONT) SelectObject( hDC, gdiFont );

	/// Generate 256 new opengl display list "names" (ID numbers), one for each ascii character
	_defaultFontListBase = glGenLists( 256 );
	wglUseFontBitmaps( hDC, 0, 255, _defaultFontListBase );

	/// Restore the previous font
	SelectObject( hDC, oldGdiFont );

	/// Delete the temporary GDI font object
	DeleteObject( gdiFont );
}


//-------------------------------------------------------------------------------------------------
// _RenderConsole
//-------------------------------------------------------------------------------------------------
void TheRenderer::_RenderConsole() const
{
	/// Balk immediately if console drawing is disabled
	if( _consoleDisplayMode == CONSOLE_DISPLAY_NONE )
		return;

	/// Save previous opengl states
	glPushAttrib( GL_COLOR_BUFFER_BIT );

	/// Get the physical pixel size of the current video mode and set a drawing coordinate system
	int viewPixelWidth = theVideoSystem.GetCurrentVidModeWidth();
	int viewPixelHeight = theVideoSystem.GetCurrentVidModeHeight();
	Vector2 bottomLeft( 0.0f, float( viewPixelHeight ) );
	Vector2 topRight( float( viewPixelWidth ), 0.0f );
	SetCoordinateSystem( bottomLeft, topRight );

	/// Calculate the pixel-height of each console row as printed onscreen
	int fontPixelHeight = DEFAULT_FONT_HEIGHT + DEFAULT_FONT_PADDING_Y;

	/// Build a rect which defines the console render area
	Vector2 areaMins( 0.0f, 0.0f );
	Vector2 areaMaxs( static_cast<float>(viewPixelWidth), static_cast<float>(viewPixelHeight) );
	if( _consoleDisplayMode == CONSOLE_DISPLAY_PARTIAL )
	{
		areaMaxs.y = float( _numConsolePartialRows * fontPixelHeight );
		areaMaxs.y += float( fontPixelHeight ) * 0.5f;
	}

	/// Calculate the number of lines of text that can fit in the console area
	int maxTextRows = viewPixelHeight / fontPixelHeight;
	int numTextRows = maxTextRows;
	if( _consoleDisplayMode == CONSOLE_DISPLAY_PARTIAL )
	{
		if( numTextRows > _numConsolePartialRows )
		{
			numTextRows = _numConsolePartialRows;
		}
	}

	/// Calculate the highest and lowest console line print numbers
	int highPrintNum = theConsole.GetHighestPrintLineNumber();
	int lowPrintNum = theConsole.GetLowestPrintLineNumber();
	int screenBottomRow = highPrintNum;
	int screenTopRow = highPrintNum - numTextRows + 1;
	if( screenTopRow < lowPrintNum )
	{
		screenTopRow = lowPrintNum;
	}

	/// Draw a translucent quad behind the console text area
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glColor4f( 0.0f, 0.0f, 0.0f, 0.5f );
	glBegin( GL_QUADS );
	{
		glVertex2f( areaMins.x, areaMins.y );
		glVertex2f( areaMaxs.x, areaMins.y );
		glVertex2f( areaMaxs.x, areaMaxs.y );
		glVertex2f( areaMins.x, areaMaxs.y );
	}
	glEnd(); // GL_QUADS

	/// Start with the lowest print number at the top of the screen
	Vector2 rasterPos( areaMins.x + 2.0f, areaMins.y );
	for( int printNum = screenTopRow; printNum <= screenBottomRow; printNum ++ )
	{
		/// Advance the drawing position downward the equivalent of 1 row height
		rasterPos.y += static_cast<float>(fontPixelHeight);

		/// Fetch the text for this line from the console system
		const string& lineText = theConsole.GetTextForPrintLineNumber( printNum );

		/// Print it on the screen; color the last row to make it stand out
		if( printNum == screenBottomRow )
		{
			XYPrintf( rasterPos.x, rasterPos.y, Rgba::ORANGE, lineText.c_str() );
		}
		else
		{
			XYPrintf( rasterPos.x, rasterPos.y, Rgba::WHITE, lineText.c_str() );
		}
	}

	/// Restore previously saved opengl states
	glPopAttrib();
}


//-------------------------------------------------------------------------------------------------
// _RenderFPSIndicator
//-------------------------------------------------------------------------------------------------
void TheRenderer::_RenderFPSIndicator() const
{
	/// Balk immediately if fps drawing is disabled
	if( !_displayFPS )
		return;

	/// Save previous opengl states
	glPushAttrib( GL_COLOR_BUFFER_BIT );

	/// Get the physical pixel size of the current video mode and set a drawing coordinate system
	int viewPixelWidth = theVideoSystem.GetCurrentVidModeWidth();
	int viewPixelHeight = theVideoSystem.GetCurrentVidModeHeight();
	Vector2 bottomLeft( 0.0f, float( viewPixelHeight ) );
	Vector2 topRight( float( viewPixelWidth ), 0.0f );
	SetCoordinateSystem( bottomLeft, topRight );

	/// Change blending modes temporarily
	if( _consoleDisplayMode == CONSOLE_DISPLAY_NONE )
	{
		glBlendFunc( GL_ONE_MINUS_DST_COLOR, GL_ZERO );
	}
	else
	{
		glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	}

	/// Draw fps counter
	Vector2 rasterPos( static_cast<float>(viewPixelWidth - 40.0f), static_cast<float>(DEFAULT_FONT_HEIGHT) );
	XYPrintf( rasterPos.x, rasterPos.y, Rgba::ORANGE, "%.0f fps", theApplication.GetFpsAsymptotic() );

	/// Restore to previous blending state
	glPopAttrib();
}


//-------------------------------------------------------------------------------------------------
// RenderScene
//-------------------------------------------------------------------------------------------------
void TheRenderer::RenderScene() const
{
	/// Check if renderer is up and active
	if( !_isInitialized )
	{
		/// If not, do nothing
		return;
	}

	/// Draw console
	_RenderConsole();

	/// Draw FPS counter
	_RenderFPSIndicator();
}


//-------------------------------------------------------------------------------------------------
// ClearScreen
//-------------------------------------------------------------------------------------------------
void TheRenderer::ClearScreen( float red, float green, float blue, float alpha ) const
{
	Rgba rgba( red, green, blue, alpha );
	ClearScreen( rgba );
}


//-------------------------------------------------------------------------------------------------
// ClearScreen
//-------------------------------------------------------------------------------------------------
void TheRenderer::ClearScreen( const Rgba& rgba ) const
{
	/// Clear the screen to the specified color and reset the z-buffer
	GLbitfield clearBits = GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT;
	glClearColor( rgba.r, rgba.g, rgba.b, rgba.a );
	glClear( clearBits );
}


//-------------------------------------------------------------------------------------------------
// ResetView
//-------------------------------------------------------------------------------------------------
void TheRenderer::ResetView() const
{
	/// Reset the viewport
	glViewport( 0, 0, theVideoSystem.GetCurrentVidModeWidth(), theVideoSystem.GetCurrentVidModeHeight() );

	/// Reset the transform matrices
	SetCoordinateSystem( Vector2::ZERO, Vector2::ONE );
}


//-------------------------------------------------------------------------------------------------
// XYPrintChars
// 
// Print a colored text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintChars( const Vector2& pos, const Rgba& rgba, const char* text, int textLength ) const
{
	/// Set the color and draw the text on the screen (using opengl display lists, one per character)
	SetColor( rgba );
	glListBase( _defaultFontListBase );
	glRasterPos2f( pos.x, pos.y );
	glCallLists( textLength, GL_UNSIGNED_BYTE, text );
}


//-------------------------------------------------------------------------------------------------
// XYPrintf( x, y )
// 
// Print a text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintf( float x, float y, const char* format, ... ) const
{
	ASSERT_MSG0( format, SUBSYSTEM_NAME"::XYPrintf() : the <format> string was NULL!" );

	/// Allocate a moderately-sized text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 256;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	int textLength = strlen( text );

	/// Draw colored text on the screen
	Vector2 pos( x, y );
	XYPrintChars( pos, Rgba::WHITE, text, textLength );
}


//-------------------------------------------------------------------------------------------------
// XYPrintf( x, y, rgba )
// 
// Print a colored text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintf( float x, float y, const Rgba& rgba, const char* format, ... ) const
{
	ASSERT_MSG0( format, SUBSYSTEM_NAME"::XYPrintf() : the <format> string was NULL!" );

	/// Allocate a moderately-sized text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 256;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	int textLength = strlen( text );

	/// Draw colored text on the screen
	Vector2 pos( x, y );
	XYPrintChars( pos, rgba, text, textLength );
}


//-------------------------------------------------------------------------------------------------
// XYPrintf( x, y, r, g, b, a )
// 
// Print a colored text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintf( float x, float y, float red, float green, float blue, float alpha, const char* format, ... ) const
{
	ASSERT_MSG0( format, SUBSYSTEM_NAME"::XYPrintf() : the <format> string was NULL!" );

	/// Allocate a moderately-sized text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 256;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	int textLength = strlen( text );

	/// Draw colored text on the screen
	Vector2 pos( x, y );
	Rgba rgba( red, green, blue, alpha );
	XYPrintChars( pos, rgba, text, textLength );
}


//-------------------------------------------------------------------------------------------------
// XYPrintf( pos )
// 
// Print a text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintf( const Vector2& pos, const char* format, ... ) const
{
	ASSERT_MSG0( format, SUBSYSTEM_NAME"::XYPrintf() : the <format> string was NULL!" );

	/// Allocate a moderately-sized text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 256;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	int textLength = strlen( text );

	/// Draw text on the screen
	XYPrintChars( pos, Rgba::WHITE, text, textLength );
}


//-------------------------------------------------------------------------------------------------
// XYPrintf( pos, rgba )
// 
// Print a colored text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintf( const Vector2& pos, const Rgba& rgba, const char* format, ... ) const
{
	ASSERT_MSG0( format, SUBSYSTEM_NAME"::XYPrintf() : the <format> string was NULL!" );

	/// Allocate a moderately-sized text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 256;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	int textLength = strlen( text );

	/// Draw colored text on the screen
	XYPrintChars( pos, rgba, text, textLength );
}


//-------------------------------------------------------------------------------------------------
// XYPrintf( pos, r, g, b, a )
// 
// Print a colored text message on the screen using the default font.
//-------------------------------------------------------------------------------------------------
void TheRenderer::XYPrintf( const Vector2& pos, float red, float green, float blue, float alpha, const char* format, ... ) const
{
	ASSERT_MSG0( format, SUBSYSTEM_NAME"::XYPrintf() : the <format> string was NULL!" );

	/// Allocate a moderately-sized text buffer to hold the args-expanded characters
	const unsigned int	TEXT_BUFFER_SIZE = 256;
	char				text[ TEXT_BUFFER_SIZE ];
	va_list				args;

	/// Write formatted string into <text>
	va_start( args, format );
	_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
	va_end( args );
	text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	int textLength = strlen( text );

	/// Draw colored text on the screen
	Rgba rgba( red, green, blue, alpha );
	XYPrintChars( pos, rgba, text, textLength );
}


//-------------------------------------------------------------------------------------------------
// ShowConsoleToggle
//-------------------------------------------------------------------------------------------------
void TheRenderer::ShowConsoleToggle()
{
	_consoleDisplayMode ++;
	_consoleDisplayMode %= CONSOLE_DISPLAY_NUMBER_OF_MODES;
}


//-------------------------------------------------------------------------------------------------
// ShowFPSToggle
//-------------------------------------------------------------------------------------------------
void TheRenderer::ShowFpsToggle()
{
	_displayFPS = !_displayFPS;
}


//-------------------------------------------------------------------------------------------------
// SetCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::SetCoordinateSystem( const Vector2& bottomLeft, const Vector2& topRight ) const
{
	Vector2 size = topRight - bottomLeft; // note, these can (correctly) be negative

	/// Reset the transform matrices
	glMatrixMode( GL_PROJECTION );
	glLoadIdentity();
	glMatrixMode( GL_MODELVIEW );
	glLoadIdentity();
	glOrtho( bottomLeft.x, topRight.x, bottomLeft.y, topRight.y, 0.0f, 1.0f );
}


//-------------------------------------------------------------------------------------------------
// TranslateCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::TranslateCoordinateSystem( const Vector2& translation )
{
	/// Translate the coordinate system
	glTranslatef( translation.x, translation.y, 0.0f );
}


//-------------------------------------------------------------------------------------------------
// TranslateCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::TranslateCoordinateSystem( const Vector2& worldReferencePoint, const Vector2& screenReferencePoint )
{
	/// Translate the coordinate system
	Vector2 translation = screenReferencePoint - worldReferencePoint;
	TranslateCoordinateSystem( translation );
}


//-------------------------------------------------------------------------------------------------
// ScaleCoordinateSystem
// 
//	Scale the coordinate system, with scaling centered around <centerPoint>.
//-------------------------------------------------------------------------------------------------
void TheRenderer::ScaleCoordinateSystem( const Vector2& scaleFactors, const Vector2& centerPoint )
{
	/// Make <centerPoint> our origin
	glTranslatef( centerPoint.x, centerPoint.y, 0.0f );

	/// Perform a 2d scale
	glScalef( scaleFactors.x, scaleFactors.y, 0.0f );

	/// Undo original move (in scaled space)
	glTranslatef( -centerPoint.x, -centerPoint.y, 0.0f );
}


//-------------------------------------------------------------------------------------------------
// ScaleCoordinateSystem
// 
//	Scale the coordinate system, with scaling centered around <centerPoint>.
//-------------------------------------------------------------------------------------------------
void TheRenderer::ScaleCoordinateSystem( const Vector2& scaleFactors )
{
	/// Perform a 2d scale
	glScalef( scaleFactors.x, scaleFactors.y, 0.0f );
}


//-------------------------------------------------------------------------------------------------
// ScaleCoordinateSystem
// 
//	Scale the coordinate system, with scaling centered around <centerPoint>.
//-------------------------------------------------------------------------------------------------
void TheRenderer::ScaleCoordinateSystem( float scaleFactor, const Vector2& centerPoint )
{
	Vector2 scaleFactors( scaleFactor, scaleFactor );
	ScaleCoordinateSystem( scaleFactors, centerPoint );
}


//-------------------------------------------------------------------------------------------------
// ScaleCoordinateSystem
// 
//	Scale the coordinate system, with scaling centered around <centerPoint>.
//-------------------------------------------------------------------------------------------------
void TheRenderer::ScaleCoordinateSystem( float scaleFactor )
{
	Vector2 scaleFactors( scaleFactor, scaleFactor );
	ScaleCoordinateSystem( scaleFactors );
}


//-------------------------------------------------------------------------------------------------
// RotateCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::RotateCoordinateSystem( float radians, const Vector2& centerPoint )
{
	/// Make <centerPoint> our origin
	glTranslatef( centerPoint.x, centerPoint.y, 0.0f );

	/// Perform a 2d rotation
	glRotatef( (float) RAD2DEG( radians ), 0.0f, 0.0f, 1.0f );

	/// Undo original move (in scaled space)
	glTranslatef( -centerPoint.x, -centerPoint.y, 0.0f );
}


//-------------------------------------------------------------------------------------------------
// RotateCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::RotateCoordinateSystem( float radians )
{
	/// Perform a 2d rotation
	glRotatef( (float) RAD2DEG( radians ), 0.0f, 0.0f, 1.0f );
}


//-------------------------------------------------------------------------------------------------
// PushCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::PushCoordinateSystem()
{
	glPushMatrix();
}


//-------------------------------------------------------------------------------------------------
// PopCoordinateSystem
//-------------------------------------------------------------------------------------------------
void TheRenderer::PopCoordinateSystem()
{
	glPopMatrix();
}


//-------------------------------------------------------------------------------------------------
// SetColor
//-------------------------------------------------------------------------------------------------
void TheRenderer::SetColor( const Rgba& rgba ) const
{
	glColor4fv( rgba.GetAsFloatArray() );
}


//-------------------------------------------------------------------------------------------------
// SetColor
//-------------------------------------------------------------------------------------------------
void TheRenderer::SetColor( float red, float green, float blue, float alpha ) const
{
	glColor4f( red, green, blue, alpha );
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRect (int, vec2, vec2, rgba, bool)
// 
// Set <setOpenglStates> to false to minimize state changes when drawing many textured
//	objects in a row; set <openglTexNum> to -1 to skip texture binding and use the
//	currently bound texture (whatever it may be).
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRect( int openglTexNum, const Vector2& bottomLeft, const Vector2& topRight, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	/// Store previous opengl states (if requested)
	if( setOpenglStates )
	{
		glPushAttrib( GL_ENABLE_BIT | GL_LIGHTING_BIT );
		glEnable( GL_TEXTURE_2D );
		glShadeModel( GL_FLAT );
	}

	/// Bind texture if a texnum was specified
	if( openglTexNum > 0 )
	{
		SetCurrentTexture( openglTexNum );
	}

	/// Draw a flat-shaded, textured quad
	SetColor( rgba );
	glBegin( GL_QUADS );
	{
		glTexCoord2f( 0.0f, 0.0f ); // bottom-left
		glVertex2f( bottomLeft.x, bottomLeft.y );
		
		glTexCoord2f( 0.0f, 1.0f ); // top-left
		glVertex2f( bottomLeft.x, topRight.y );
		
		glTexCoord2f( 1.0f, 1.0f ); // top-right
		glVertex2f( topRight.x, topRight.y );
		
		glTexCoord2f( 1.0f, 0.0f ); // bottom-right
		glVertex2f( topRight.x, bottomLeft.y );
	}
	glEnd(); // GL_QUADS
	/// Restore previous opengl states (if requested)
	if( setOpenglStates )
	{
		glPopAttrib();
	}
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRect
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRect( const Texture& texture, const Vector2& bottomLeft, const Vector2& topRight, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	DrawTexturedRect( texture.GetGLTexNum(), bottomLeft, topRight, rotationDegrees, scale, rgba, setOpenglStates );
}



//-------------------------------------------------------------------------------------------------
// DrawTexturedRect
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRect( int openglTexNum, const Vector2& bottomLeft, const Vector2& topRight, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	/// Rotate coordinate system if requested
	if( (rotationDegrees != 0.0f) || (scale != Vector2::ONE) )
	{
		PushCoordinateSystem();

		/// Calculate the center point
		Vector2 center = (bottomLeft + topRight) * 0.5f;
		RotateCoordinateSystem( DEG2RAD( rotationDegrees ), center );
		ScaleCoordinateSystem( scale, center );
	}


	/// Store previous opengl states (if requested)
	if( setOpenglStates )
	{
		glPushAttrib( GL_ENABLE_BIT | GL_LIGHTING_BIT );
		glEnable( GL_TEXTURE_2D );
		glShadeModel( GL_FLAT );
	}

	/// Bind texture if a texnum was specified
	if( openglTexNum > 0 )
	{
		SetCurrentTexture( openglTexNum );
	}

	/// Draw a flat-shaded, textured quad
	SetColor( rgba );
	glBegin( GL_QUADS );
	{
		glTexCoord2f( texCoordTopLeft.x, texCoordTopLeft.y ); // bottom-left
		glVertex2f( bottomLeft.x, bottomLeft.y );
		
		glTexCoord2f( texCoordTopLeft.x, texCoordBottomRight.y ); // top-left
		glVertex2f( bottomLeft.x, topRight.y );
		
		glTexCoord2f( texCoordBottomRight.x, texCoordBottomRight.y ); // top-right
		glVertex2f( topRight.x, topRight.y );
		
		glTexCoord2f( texCoordBottomRight.x, texCoordTopLeft.y ); // bottom-right
		glVertex2f( topRight.x, bottomLeft.y );
	}
	glEnd(); // GL_QUADS
	/// Restore previous opengl states (if requested)
	if( setOpenglStates )
	{
		glPopAttrib();
	}

	/// Un-rotate coordinate system
	if( (rotationDegrees != 0.0f) || (scale != Vector2::ONE) )
	{
		PopCoordinateSystem();
	}
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRect
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRect( const Texture& texture, const Vector2& bottomLeft, const Vector2& topRight, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	DrawTexturedRect( texture.GetGLTexNum(), bottomLeft, topRight, texCoordTopLeft, texCoordBottomRight, rotationDegrees, scale, rgba, setOpenglStates );
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRectCentered (int, vec2, vec2 rgba, bool)
// 
// Set <setOpenglStates> to false to minimize state changes when drawing many textured
//	objects in a row; set <openglTexNum> to -1 to skip texture binding and use the
//	currently bound texture (whatever it may be).
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRectCentered( int openglTexNum, const Vector2& centerPoint, const Vector2& rectSize, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	Vector2 rectHalfSize = 0.5f * rectSize;
	Vector2 bottomLeft	= centerPoint - rectHalfSize;
	Vector2 topRight	= centerPoint + rectHalfSize;

	DrawTexturedRect( openglTexNum, bottomLeft, topRight, rotationDegrees, scale, rgba, setOpenglStates );
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRectCentered (int, vec2, vec2 rgba, bool)
// 
// Set <setOpenglStates> to false to minimize state changes when drawing many textured
//	objects in a row; set <openglTexNum> to -1 to skip texture binding and use the
//	currently bound texture (whatever it may be).
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRectCentered( const Texture& texture, const Vector2& centerPoint, const Vector2& rectSize, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	DrawTexturedRectCentered( texture.GetGLTexNum(), centerPoint, rectSize, rotationDegrees, scale, rgba, setOpenglStates );
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRectCentered
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRectCentered( int openglTexNum, const Vector2& centerPoint, const Vector2& rectSize, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	Vector2 rectHalfSize = 0.5f * rectSize;
	Vector2 bottomLeft	= centerPoint - rectHalfSize;
	Vector2 topRight	= centerPoint + rectHalfSize;

	DrawTexturedRect( openglTexNum, bottomLeft, topRight, texCoordTopLeft, texCoordBottomRight, rotationDegrees, scale, rgba, setOpenglStates );
}


//-------------------------------------------------------------------------------------------------
// DrawTexturedRectCenteredTexCoords
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawTexturedRectCentered( const Texture& texture, const Vector2& centerPoint, const Vector2& rectSize, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees, const Vector2& scale, const Rgba& rgba, bool setOpenglStates )
{
	DrawTexturedRectCentered( texture.GetGLTexNum(), centerPoint, rectSize, texCoordTopLeft, texCoordBottomRight, rotationDegrees, scale, rgba, setOpenglStates );
}



//-------------------------------------------------------------------------------------------------
// SetCurrentTexture
//-------------------------------------------------------------------------------------------------
void TheRenderer::SetCurrentTexture( int openglTexNum, bool forceBind )
{
	/// Check if <forceBind> was specified OR this texNum differs from the current
	if( forceBind || (openglTexNum != _currentTexNum) )
	{
		/// Bind the texture and remember it as the "current" texNum
		glBindTexture( GL_TEXTURE_2D, openglTexNum );
		_currentTexNum = openglTexNum;
	}
}


//-------------------------------------------------------------------------------------------------
// SetCurrentTexture
//-------------------------------------------------------------------------------------------------
void TheRenderer::SetCurrentTexture( const Texture& texture, bool forceBind )
{
	SetCurrentTexture( texture.GetGLTexNum(), forceBind );
}


//-------------------------------------------------------------------------------------------------
// DrawVector
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawVector( const Vector2& start, const Vector2& end, const Rgba& startColor, const Rgba& endColor )
{
	glPushAttrib( GL_ENABLE_BIT | GL_LIGHTING_BIT | GL_COLOR_BUFFER_BIT );

	glDisable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	glShadeModel( GL_SMOOTH );

	glBegin( GL_LINES );
	{
		SetColor( startColor );
		glVertex2f( start.x, start.y );
		SetColor( endColor );
		glVertex2f( end.x, end.y );
	}
	glEnd();

	glPopAttrib();
}


//-------------------------------------------------------------------------------------------------
// DrawVector
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawVector( const LineSegment2& lineSegment, const Rgba& startColor, const Rgba& endColor )
{
	DrawVector( lineSegment.start, lineSegment.start + lineSegment.displacement, startColor, endColor );
}


//-------------------------------------------------------------------------------------------------
// DrawAABB2
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawAABB2( const AABB2& bounds, const Rgba& color )
{
	glPushAttrib( GL_ENABLE_BIT | GL_LIGHTING_BIT | GL_COLOR_BUFFER_BIT );

	glDisable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	SetColor( color );

	glBegin( GL_LINE_LOOP );
	{
		glVertex2f( bounds.mins.x, bounds.mins.y );
		glVertex2f( bounds.maxs.x, bounds.mins.y );
		glVertex2f( bounds.maxs.x, bounds.maxs.y );
		glVertex2f( bounds.mins.x, bounds.maxs.y );
	}
	glEnd();

	glPopAttrib();
}


//-------------------------------------------------------------------------------------------------
// DrawCircle
// 
// Specify -1 for <numSides> to let the function determine the number of sides
//	to use when drawing the circle based on its size.
//-------------------------------------------------------------------------------------------------
void TheRenderer::DrawCircle( const Vector2& centerPoint, float radius, const Rgba& rgba, int numSides )
{
	glPushAttrib( GL_ENABLE_BIT | GL_LIGHTING_BIT | GL_COLOR_BUFFER_BIT );

	glDisable( GL_TEXTURE_2D );
	glEnable( GL_BLEND );
	glShadeModel( GL_SMOOTH );
	SetColor( rgba );

	if( numSides == -1 )
	{
		numSides = 8 + (int) (radius * 0.2f);
		if( numSides < 12 )
			numSides = 12;
	}

	glBegin( GL_LINE_LOOP );
	for( int i = 0; i < numSides; ++ i )
	{
		float f = (float) i / (float) numSides;
		float radians = f * (float) TWO_PI;
		Vector2 point;
		point.SetFromCosSinRadians( radians );
		point *= radius;
		point += centerPoint;
		glVertex2f( point.x, point.y );
	}
	glEnd();

	glPopAttrib();
}



