//-------------------------------------------------------------------------------------------------
// TheRenderer.hpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#ifndef _include_TheRenderer_
#define _include_TheRenderer_


//-------------------------------------------------------------------------------------------------
// Included header files
//-------------------------------------------------------------------------------------------------
#include "engineCommon.hpp"
#include "Rgba.hpp"
#include "Vector2.hpp"
#include "AABB2.hpp"
#include "LineSegment2.hpp"
#include "Image.hpp"
#include "Texture.hpp"


///////////////////////////////////////////////////////////////////////////////////////////////////
//
//		TheRenderer
//
///////////////////////////////////////////////////////////////////////////////////////////////////
class TheRenderer
{
private:
	enum
	{
		DEFAULT_FONT_HEIGHT			= 16,
		DEFAULT_FONT_PADDING_Y		= -2,
	};

	enum
	{
		CONSOLE_DISPLAY_NONE		= 0,
		CONSOLE_DISPLAY_PARTIAL		= 1,
		CONSOLE_DISPLAY_FULL		= 2,
		CONSOLE_DISPLAY_NUMBER_OF_MODES,
	};

	/// Static members
	static bool _isInitialized;

	/// Private members
	float _zBufferDepthMin;
	float _zBufferDepthMax;
	int _defaultFontListBase;	// opengl display list "name" (ID number) where ascii font starts
	int _consoleDisplayMode;
	bool _displayFPS;
	int _numConsolePartialRows;
	int _currentTexNum;

	/// Private methods
	void _InitOpenGLStates() const;
	void _InitFonts();
	void _RenderConsole() const;
	void _RenderFPSIndicator() const;

public:
	/// Construction, initialization, and singleton methods
	TheRenderer(); // constructor
	~TheRenderer(); // destructor

	void Startup();
	void Shutdown();

	/// Public methods
	void ClearScreen( float red, float green, float blue, float alpha ) const;
	void ClearScreen( const Rgba& rgba ) const;
	void ResetView() const;
	void RenderScene() const;
	void XYPrintf( float x, float y, const char* format, ... ) const;
	void XYPrintf( float x, float y, const Rgba& rgba, const char* format, ... ) const;
	void XYPrintf( float x, float y, float red, float green, float blue, float alpha, const char* format, ... ) const;
	void XYPrintf( const Vector2& pos, const char* format, ... ) const;
	void XYPrintf( const Vector2& pos, const Rgba& rgba, const char* format, ... ) const;
	void XYPrintf( const Vector2& pos, float red, float green, float blue, float alpha, const char* format, ... ) const;
	void XYPrintChars( const Vector2& pos, const Rgba& rgba, const char* text, int textLength ) const;
	void ShowConsoleToggle();
	void ShowFpsToggle();

	void SetCoordinateSystem( const Vector2& bottomLeft, const Vector2& topRight ) const;
	void TranslateCoordinateSystem( const Vector2& translation );
	void TranslateCoordinateSystem( const Vector2& worldReferencePoint, const Vector2& screenReferencePoint );
	void ScaleCoordinateSystem( const Vector2& scaleFactors, const Vector2& centerPoint );
	void ScaleCoordinateSystem( float scaleFactor, const Vector2& centerPoint );
	void ScaleCoordinateSystem( const Vector2& scaleFactors );
	void ScaleCoordinateSystem( float scaleFactor );
	void RotateCoordinateSystem( float radians, const Vector2& centerPoint );
	void RotateCoordinateSystem( float radians );
	void PushCoordinateSystem();
	void PopCoordinateSystem();

	void SetColor( const Rgba& rgba ) const;
	void SetColor( float red, float green, float blue, float alpha = 1.0f ) const;
	void DrawTexturedRect( int openglTexNum, const Vector2& bottomLeft, const Vector2& topRight, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRect( const Texture& texture, const Vector2& bottomLeft, const Vector2& topRight, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRect( int openglTexNum, const Vector2& bottomLeft, const Vector2& topRight, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRect( const Texture& texture, const Vector2& bottomLeft, const Vector2& topRight, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRectCentered( int openglTexNum, const Vector2& centerPoint, const Vector2& rectSize, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRectCentered( const Texture& texture, const Vector2& centerPoint, const Vector2& rectSize, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRectCentered( int openglTexNum, const Vector2& centerPoint, const Vector2& rectSize, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void DrawTexturedRectCentered( const Texture& texture, const Vector2& centerPoint, const Vector2& rectSize, const Vector2& texCoordTopLeft, const Vector2& texCoordBottomRight, float rotationDegrees = 0.0f, const Vector2& scale = Vector2::ONE, const Rgba& rgba = Rgba::WHITE, bool setOpenglStates = true );
	void SetCurrentTexture( int openglTexNum, bool forceBind = false );
	void SetCurrentTexture( const Texture& texture, bool forceBind = false );
	void DrawVector( const Vector2& start, const Vector2& end, const Rgba& startColor = Rgba::WHITE, const Rgba& endColor = Rgba::WHITE );
	void DrawVector( const LineSegment2& lineSegment, const Rgba& startColor = Rgba::WHITE, const Rgba& endColor = Rgba::WHITE );
	void DrawCircle( const Vector2& centerPoint, float radius, const Rgba& rgba = Rgba::WHITE, int numSides = -1 );
	void DrawAABB2( const AABB2& bounds, const Rgba& color = Rgba::WHITE );
};



//-------------------------------------------------------------------------------------------------
// Global singleton instance
//-------------------------------------------------------------------------------------------------
extern TheRenderer theRenderer;


#endif // _include_TheRenderer_
