//-------------------------------------------------------------------------------------------------
// TheVideoSystem.cpp (engine)
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//-------------------------------------------------------------------------------------------------
#include "pch.hpp"
#include "TheVideoSystem.hpp"
#include "TheConsole.hpp"
#include "TheApplication.hpp"
#include "engineAssert.hpp"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Subsystem name macro (for assertions, etc)
//-------------------------------------------------------------------------------------------------
#define SUBSYSTEM_NAME "TheVideoSystem"


//-------------------------------------------------------------------------------------------------
// Class constants
//-------------------------------------------------------------------------------------------------
const DWORD
	TheVideoSystem::WIN_STYLE_FLAGS_WINDOWED		= ( WS_CAPTION | WS_SYSMENU | WS_OVERLAPPED ),
	TheVideoSystem::WIN_STYLE_FLAGS_FULLSCREEN		= ( WS_POPUP ),
	TheVideoSystem::WIN_STYLE_EX_FLAGS_WINDOWED		= ( WS_EX_APPWINDOW ),
	TheVideoSystem::WIN_STYLE_EX_FLAGS_FULLSCREEN	= ( WS_EX_APPWINDOW );


//-------------------------------------------------------------------------------------------------
// Static member initialization
//-------------------------------------------------------------------------------------------------
STATIC bool				TheVideoSystem::_isInitialized		= false;
STATIC HWND				TheVideoSystem::_hWnd				= NULL;


//-------------------------------------------------------------------------------------------------
// Forward declarations, prototypes, and externs
//-------------------------------------------------------------------------------------------------
LRESULT CALLBACK WinProc(	HWND	hWnd,		// handle for this window
							UINT	uMsg,		// the type of message being received
							WPARAM	wParam,		// additional message data (optional)
							LPARAM	lParam );	// additional message data (optional)


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
TheVideoSystem::TheVideoSystem()
:
OPENGL_WINDOW_CLASS_NAME( "OpenGL GameEngine Window Class" ),
_defaultWidth( 640 ),
_defaultHeight( 480 ),
_defaultBpp( 32 ),
_defaultFrequency( 60 ),
_defaultZBitDepth( 16 ),
_defaultFlags( 0 ),
_numVidModes( 0 ),
_currentVidMode( 0 ),
_hDC( NULL ),
_hRC( NULL ),
_hDesktopWnd( NULL ),
_showGDIStatusMessages( true )
{
	/// Initialize complex members
	memset( _vidModes, 0, sizeof( *_vidModes ) );
	memset( &_clientRect, 0, sizeof( _clientRect ) );
	memset( &_clientOffset, 0, sizeof( _clientOffset ) );
	memset( &_windowRect, 0, sizeof( _windowRect ) );
	memset( &_desktopRect, 0, sizeof( _desktopRect ) );
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
TheVideoSystem::~TheVideoSystem()
{
	/// Make sure the subsystem was shut down before being destroyed
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was destroyed while running; Shutdown() should be called first!" );

	/// Free the video mode table
	for( unsigned int i = 0; i < _numVidModes; ++ i )
	{
		delete _vidModes[ i ];
		_vidModes[ i ] = NULL;
	}
	_numVidModes = 0;
}


//-------------------------------------------------------------------------------------------------
// Startup
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::Startup()
{
	theConsole.Printf( "::: Initialization : Video System (starting)\n" );

	/// Make sure the subsystem doesn't get initialized twice
	VALIDATE_MSG0( !_isInitialized, SUBSYSTEM_NAME" was initialized twice; either a second "SUBSYSTEM_NAME" was started or Startup() was called on an already-running system." );

	/// Get desktop characteristics
	_GetDesktopCharacteristics();

	/// Populate the list of fullscreen video modes available
	_DetermineAvailableVideoModes();

	/// Select the video mode closest to the defaults
	_SetModeClosestToDefaults();

	/// Create a window and a rendering context
	_CreateRenderingWindow();

	/// Set initialization complete
	_isInitialized = true;
	theConsole.Printf( "::: Initialization : Video System (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// Shutdown
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::Shutdown()
{
	theConsole.Printf( "::: Shutdown : Video System (starting)\n" );

	/// Restore the desktop
	ChangeDisplaySettings( NULL, 0 );

	/// Hide the application window and open a message box
	ShowWindow( _hWnd, SW_HIDE ); // hide the window (forever!)

	/// De-initialize the system
	_isInitialized = false;
	theConsole.Printf( "::: Shutdown : Video System (finished)\n" );
}


//-------------------------------------------------------------------------------------------------
// _DetermineAvailableVideoModes
// 
// Builds a list of all available display modes and returns the total number of them.
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::_DetermineAvailableVideoModes()
{
	_numVidModes = 0;

	/// Create some windowed display modes of standard sizes
	_CreateWindowedDisplayMode(  320,  240, 32 );
	_CreateWindowedDisplayMode(  400,  300, 32 );
	_CreateWindowedDisplayMode(  512,  384, 32 );
	_CreateWindowedDisplayMode(  640,  480, 32 );
	_CreateWindowedDisplayMode(  800,  600, 32 );
	_CreateWindowedDisplayMode( 1024,  768, 32 );
	_CreateWindowedDisplayMode( 1152,  864, 32 );
	_CreateWindowedDisplayMode( 1280,  960, 32 );
	_CreateWindowedDisplayMode( 1600, 1200, 32 );

	/// Enumerate through all available fullscreen display modes
	while( _numVidModes < MAX_VID_MODES )
	{
		/// Allocate memory for the next fullscreen mode (assuming there is one)
		_vidModes[ _numVidModes ] = new VidModeDesc;
		_vidModes[ _numVidModes ]->flags = VIDMODE_FLAG_FULLSCREEN;

		/// Attempt to query the next display mode in the list of available modes
		BOOL success = EnumDisplaySettings( NULL, _numVidModes, &_vidModes[ _numVidModes ]->devMode );
		if( !success )
		{
			/// Free and clear the last entry; it will not be needed after all
			delete _vidModes[ _numVidModes ];
			_vidModes[ _numVidModes ] = NULL;

			/// Break out of the loop
			break;
		}

		_numVidModes ++;
	}

	/// Check if the vidmode table overflowed
	VALIDATE( _numVidModes < MAX_VID_MODES );

	/// Check if the table is empty
	VALIDATE_MSG0( _numVidModes > 0, SUBSYSTEM_NAME" : couldn't enumerate ANY valid video modes!" );

	/// Report available display modes 
	ListVidModes();
	theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "VideoSystem : found %d available display mode(s)\n", _numVidModes );

	return( _numVidModes );
}


//-------------------------------------------------------------------------------------------------
// ListVidModes
// 
// DESCRIPTION
//	Lists all available display modes in the console.
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::ListVidModes() const
{
	const char WINDOWED_STRING[] = "windowed";
	const char FULLSCREEN_STRING[] = "fullscreen";

	/// Loop through list of display modes and print each to the console
	theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "VideoSystem : found %d display mode(s)%c\n", _numVidModes, _numVidModes ? ':' : '.' );
	for( unsigned int i = 0; i < _numVidModes; i ++ )
	{
		/// Get some convenience references and description string(s)
		VidModeDesc& vidMode = *_vidModes[ i ];
		DEVMODE& devMode = vidMode.devMode;
		const char* windowedDesc = FULLSCREEN_STRING;
		if( IsVidModeWindowed( i ) )
			windowedDesc = WINDOWED_STRING;

		/// Print display mode properties
		if( devMode.dmDisplayFrequency )
		{
			theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "  MODE %3d :  %4d x %4d (%3d bpp, %3d Hz, %s)\n", i,
				devMode.dmPelsWidth,
				devMode.dmPelsHeight,
				devMode.dmBitsPerPel,
				devMode.dmDisplayFrequency,
				windowedDesc );
		}
		else
		{
			theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "  MODE %3d :  %4d x %4d (%3d bpp, ??? Hz, %s)\n", i,
				devMode.dmPelsWidth,
				devMode.dmPelsHeight,
				devMode.dmBitsPerPel,
				windowedDesc );
		}
	}
}


//-------------------------------------------------------------------------------------------------
// _GetDesktopCharacteristics
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_GetDesktopCharacteristics()
{
	/// Get a handle to the desktop window
	_hDesktopWnd = GetDesktopWindow();

	/// Get the desktop window
	BOOL success = GetWindowRect( _hDesktopWnd, &_desktopRect );
	VALIDATE_MSG0( success, "TheVideoSystem::_GetDesktopCharacteristics : GetWindowRect() failed!" );
}


//-------------------------------------------------------------------------------------------------
// SwapBuffers
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::SwapBuffers() const
{
	ASSERT_MSG0( _isInitialized, SUBSYSTEM_NAME"::SwapBuffers() was called while "SUBSYSTEM_NAME" was uninitialized; call Startup() first." );

	::SwapBuffers( _hDC );
}


//-------------------------------------------------------------------------------------------------
// _CreateRenderingWindow
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_CreateRenderingWindow()
{
	/// Get a convenience reference
	theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "VideoSystem : Creating a rendering window for mode: %s\n", GetVidModeDescription( _currentVidMode ).c_str() );
	VidModeDesc& vidMode = *_vidModes[ _currentVidMode ];

	/// Set up the window class description & register it
	WNDCLASS wndClass; // new window class properties
	_FillWindowClassDescription( wndClass );
	ATOM atom = RegisterClass( &wndClass );
	VALIDATE_MSG0( atom != 0, "VideoSystem : RegisterClass() failed!" );

	/// Attempt to go fullscreen if the current video mode dictates it
	if( !IsCurrentVidModeWindowed() )
	{
		/// Try to set selected mode (fullscreen).  CDS_FULLSCREEN gets rid of start bar, etc.
		LONG result = ChangeDisplaySettings( &vidMode.devMode, CDS_FULLSCREEN );
		VALIDATE_MSG1( result == DISP_CHANGE_SUCCESSFUL, SUBSYSTEM_NAME" : failed to create a fullscreen mode matching \"%s\"", GetVidModeDescription( _currentVidMode ).c_str() );
	}

	/// Set up bitflags for <style> and <style_ex> appropriate to the desired video mode
	DWORD style, style_ex;
	_SetStyleFlagsForVidMode( style, style_ex, _currentVidMode );

	/// Set up the client rect and window rect for this video mode
	_SetClientAndWindowRectForVidMode( _currentVidMode );

	/// Create the window
	_hWnd =  CreateWindowEx(
		style_ex,								// the extended style of the new window class
		wndClass.lpszClassName,					// name of the new window class (created above)
		theApplication.GetName().c_str(),		// title of window == name of the application (set by game)
		style,									// the window-style we chose above
		_windowRect.left,						// x-position of upper-left corner of window
		_windowRect.top,						// y-position of upper-left corner of window
		_windowRect.right - _windowRect.left,	// calculated window width
		_windowRect.bottom - _windowRect.top,	// calculated window height
		NULL,									// no parent window
		NULL,									// no windows-style menus
		theApplication.GetHInstance(),			// copy of the instance handle of the application
		NULL );									// don't pass any special info in the WM.CREATE message
	VALIDATE_MSG0( _hWnd != NULL, "TheVideoSystem::_CreateRenderingWindow : CreateWindowEx() failed!" );

	/// Initialize the window's initial display state
	ShowWindow( _hWnd, SW_SHOW );		// make the window visible
	SetForegroundWindow( _hWnd );		// put window in foreground
	SetFocus( _hWnd );					// assigns keyboard / input focus to this window

	/// Set up the desired pixel format
	PIXELFORMATDESCRIPTOR pixelFormatDesc; // color/z/alpha/stencil/accum properties
	_FillPixelFormatDescriptor( pixelFormatDesc );

	/// Create a rendering context appropriate to this pixel format preference
	_CreateRenderingContextForWindow( _hWnd, pixelFormatDesc );

	theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "VideoSystem : Successfully created a rendering window.\n" );
}


//-------------------------------------------------------------------------------------------------
// _FillWindowClassDescription
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_FillWindowClassDescription( WNDCLASS& wndClass )
{
	/// Get program icon and cursor handles
	HICON myIcon = LoadIcon( NULL, IDI_APPLICATION ); // use default icon
	HCURSOR myCursor = LoadCursor( NULL, IDC_ARROW ); // default cursor

	/// Fill in the contents of <wndClass>, the window class description
	memset( &wndClass, 0, sizeof( wndClass ) );
	wndClass.style			= CS_HREDRAW | // redraw if moved horizontally
								CS_VREDRAW | // ...or vertically
								CS_OWNDC; // acquire private DC for program
	wndClass.lpfnWndProc	= (WNDPROC) WinProc; // assign a message-handler
	wndClass.hInstance		= theApplication.GetHInstance();
	wndClass.hIcon			= myIcon;
	wndClass.hCursor		= myCursor;
	wndClass.lpszClassName	= OPENGL_WINDOW_CLASS_NAME; // name this new window class
}


//-------------------------------------------------------------------------------------------------
// _SetStyleFlagsForVidMode
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_SetStyleFlagsForVidMode( DWORD& style, DWORD& style_ex, const int vidModeNum )
{
	if( IsVidModeWindowed( vidModeNum ) )
	{
		style		= WIN_STYLE_FLAGS_WINDOWED;
		style_ex	= WIN_STYLE_EX_FLAGS_WINDOWED;
	}
	else // fullscreen
	{
		style		= WIN_STYLE_FLAGS_FULLSCREEN;
		style_ex	= WIN_STYLE_EX_FLAGS_FULLSCREEN;
	}
}


//-------------------------------------------------------------------------------------------------
// _SetClientAndWindowRectForVidMode
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_SetClientAndWindowRectForVidMode( const int vidMode )
{
	/// Get a convenience reference to the current video mode
	DEVMODE& devMode = _vidModes[ vidMode ]->devMode;

	/// Set up the client area
	_clientRect.top = 0;
	_clientRect.left = 0;
	_clientRect.right = devMode.dmPelsWidth;
	_clientRect.bottom = devMode.dmPelsHeight;

	/// Set up the window area
	_windowRect = _clientRect;
	if( IsVidModeWindowed( vidMode ) )
	{
		/// Resize the window area to accomodate the exact client size specified
		AdjustWindowRectEx( &_windowRect, WIN_STYLE_FLAGS_WINDOWED, FALSE, WIN_STYLE_EX_FLAGS_WINDOWED );
	}

	/// Move window to be onscreen, if possible
	_ClampWindowRectToDesktop( false );
}


//-------------------------------------------------------------------------------------------------
// _FillPixelFormatDescriptor
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_FillPixelFormatDescriptor( PIXELFORMATDESCRIPTOR& pfd )
{
	/// Set up the pixel format flags
	DWORD flags = PFD_DRAW_TO_WINDOW | PFD_SUPPORT_OPENGL | PFD_DOUBLEBUFFER;

	/// Fill in the contents of <pfd>, the pixel format descriptor
	memset( &pfd, 0, sizeof( pfd ) );
	pfd.nSize			= sizeof( pfd ); // size of this struct
	pfd.nVersion		= 1;
	pfd.dwFlags			= flags;
	pfd.iPixelType		= PFD_TYPE_RGBA; // pixels are in an RGBA configuration
	pfd.cColorBits		= (BYTE) _vidModes[ _currentVidMode ]->devMode.dmBitsPerPel;
	pfd.cDepthBits		= (BYTE) _defaultZBitDepth; // z-buffer bit-depth
	pfd.cAccumBits		= 0; // no accumulation buffer
	pfd.cStencilBits	= 0; // no stencil buffer
}


//-------------------------------------------------------------------------------------------------
// _CreateRenderingContextForWindow
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_CreateRenderingContextForWindow( const HWND hWnd, PIXELFORMATDESCRIPTOR& pixelFormatDesc )
{
	BOOL		success;
	int			pixelFormatNum; // pixel format # that was found

	/// Get a DC (display device context) for the display
	_hDC = GetDC( hWnd );
	VALIDATE_MSG0( _hDC != NULL, "TheVideoSystem::_CreateRenderingContextForWindow : GetDC() failed!" );

	/// Get a matching pixel format
	UpdateGDIWaitPlaque( "Choosing a pixel format..." );
	pixelFormatNum = ChoosePixelFormat( _hDC, &pixelFormatDesc );
	VALIDATE_MSG0( pixelFormatNum != 0, "TheVideoSystem::_CreateRenderingContextForWindow : ChoosePixelFormat() failed!" );

	/// Switch to the new pixel format
	UpdateGDIWaitPlaque( "Setting the pixel format..." );
	success = SetPixelFormat( _hDC, pixelFormatNum, &pixelFormatDesc );
	VALIDATE_MSG0( success, "TheVideoSystem::_CreateRenderingContextForWindow : SetPixelFormat() failed!" );

	/// Get a RC (rendering context) for this DC
	UpdateGDIWaitPlaque( "Creating a rendering context..." );
	_hRC = wglCreateContext( _hDC );
	VALIDATE_MSG0( _hRC != NULL, "TheVideoSystem::_CreateRenderingContextForWindow : wglCreateContext() failed!" );

	/// Make this new RC current
	UpdateGDIWaitPlaque( "Making rendering context current..." );
	success = wglMakeCurrent( _hDC, _hRC );
	VALIDATE_MSG0( _hRC != NULL, "TheVideoSystem::_CreateRenderingContextForWindow : wglMakeCurrent() failed!" );

	UpdateGDIWaitPlaque( NULL );
}



//-------------------------------------------------------------------------------------------------
// IsVidModeWindowed
//-------------------------------------------------------------------------------------------------
const bool TheVideoSystem::IsVidModeWindowed( const int vidMode ) const
{
	VidModeDesc& vidModeDesc = *_vidModes[ vidMode ];

	if( vidModeDesc.flags & VIDMODE_FLAG_FULLSCREEN )
		return false;

	return true;
}


//-------------------------------------------------------------------------------------------------
// IsCurrentVidModeWindowed
//-------------------------------------------------------------------------------------------------
const bool TheVideoSystem::IsCurrentVidModeWindowed() const
{
	int currentVidMode = GetCurrentVidMode();
	return IsVidModeWindowed( currentVidMode );
}


//-------------------------------------------------------------------------------------------------
// _ClampWindowRectToDesktop
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::_ClampWindowRectToDesktop( const bool forceResize )
{
	RECT windowRectBeforeMoving = _windowRect;

	/// Move the window up so that the bottom fits
	if( _windowRect.bottom > _desktopRect.bottom )
	{
		_windowRect.top -= (_windowRect.bottom - _desktopRect.bottom);
		_windowRect.bottom = _desktopRect.bottom;
	}

	/// Move the window left so that the right fits
	if( _windowRect.right > _desktopRect.right )
	{
		_windowRect.left -= (_windowRect.right - _desktopRect.right);
		_windowRect.right = _desktopRect.right;
	}

	/// Move the window down so that the top fits
	if( _windowRect.top < _desktopRect.top )
	{
		if( !forceResize )
			_windowRect.bottom += (_desktopRect.top - _windowRect.top);
		_windowRect.top = _desktopRect.top;
	}

	/// Move the window right so that the left fits
	if( _windowRect.left < _desktopRect.left )
	{
		if( !forceResize )
			_windowRect.right += (_desktopRect.left - _windowRect.left);
		_windowRect.left = _desktopRect.left;
	}

	_clientOffset.x = _windowRect.left - windowRectBeforeMoving.left;
	_clientOffset.y = _windowRect.top - windowRectBeforeMoving.top;
}


//-------------------------------------------------------------------------------------------------
// ConvertScreenCoordsToClientCoords
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::ConvertScreenCoordsToClientCoords( const int screenX, const int screenY, int& clientX, int& clientY )
{
	clientX = screenX - _clientOffset.x;
	clientY = screenY - _clientOffset.y;
}


//-------------------------------------------------------------------------------------------------
// ConvertClientCoordsToScreenCoords
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::ConvertClientCoordsToScreenCoords( const int clientX, const int clientY, int& screenX, int& screenY )
{
	screenX = clientX + _clientOffset.x;
	screenY = clientY + _clientOffset.y;
}


//-------------------------------------------------------------------------------------------------
// UpdateGDIWaitPlaque
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::UpdateGDIWaitPlaque( const char* const format, ... )
{
	const unsigned int TEXT_BUFFER_SIZE = 2048;
	va_list			args;
	char			text[ TEXT_BUFFER_SIZE ];
	HBRUSH			fillBrush;
	HGDIOBJ			textFont;
	HGDIOBJ			prevFont;
	int				xCenter;
	int				yCenter;

	if( format )
	{
		/// Write formatted string into <text>
		va_start( args, format );
		_vsnprintf( text, TEXT_BUFFER_SIZE, format, args );
		va_end( args );
		text[ TEXT_BUFFER_SIZE - 1 ] = '\0';
	}
	else
	{
		text[ 0 ] = '\0';
	}

	xCenter = (_clientRect.left + _clientRect.right) / 2;
	yCenter = (_clientRect.top + _clientRect.bottom) / 2;
	if( text )
		xCenter -= (8 * strlen( text )) / 2;

	SetBkMode( _hDC, TRANSPARENT ); 
	fillBrush = CreateSolidBrush( 0x00440000 );
	FillRect( _hDC, &_clientRect, fillBrush );

	if( text && _showGDIStatusMessages )
	{
		prevFont = GetCurrentObject( _hDC, OBJ_FONT );
		SetTextAlign( _hDC, TA_LEFT | TA_TOP | TA_NOUPDATECP );
		textFont = CreateFont(
			20,							// height of font
			10,							// average character width
			0,							// angle of escapement
			0,							// base-line orientation angle
			200,						// font weight
			FALSE,						// italic attribute option
			FALSE,						// underline attribute option
			FALSE,						// strikeout attribute option
			ANSI_CHARSET,				// character set identifier
			OUT_TT_PRECIS,				// output precision
			CLIP_DEFAULT_PRECIS,		// clipping precision
			PROOF_QUALITY,				// output quality
			DEFAULT_PITCH | FF_MODERN,	// pitch and family
			"Tahoma" );					// typeface name

		SelectObject( _hDC, textFont );
		SetTextColor( _hDC, 0x00ff8888 );
		TextOut( _hDC, xCenter, yCenter, text, strlen( text ) );
		SelectObject( _hDC, prevFont );
		DeleteObject( textFont );
	}

	DeleteObject( fillBrush );
	if( format )
	{
		theConsole.Printf( CHAN_VIDEO | CHAN_LOW, "GDI Plaque : %s\n", text );
	}
}


//-------------------------------------------------------------------------------------------------
// SetDefaultVidModeProperties
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::SetDefaultVidModeProperties(
	const int width, const int height, const int bpp, const int frequency, const unsigned int flags )
{
	_defaultWidth		= width;
	_defaultHeight		= height;
	_defaultBpp			= bpp;
	_defaultFrequency	= frequency;
	_defaultFlags		= flags;
}


//-------------------------------------------------------------------------------------------------
// _SetModeClosestToDefaults
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::_SetModeClosestToDefaults()
{
	int vidMode = SetCurrentVidModeByProperties( _defaultWidth, _defaultHeight,
		_defaultBpp, _defaultFrequency, _defaultFlags );

	return vidMode;
}


//-------------------------------------------------------------------------------------------------
// SetCurrentVidModeByProperties
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::SetCurrentVidModeByProperties(
	const int width, const int height, const int bpp, const int frequency, const unsigned int flags )
{
	int vidMode;

	/// Find the video mode most closely matching the one requested
	vidMode = _GetVidModeClosestToProperties( width, height, bpp, frequency, flags );

	/// Make it the active mode
	SetCurrentVidMode( vidMode );

	return vidMode;
}


//-------------------------------------------------------------------------------------------------
// _GetVidModeClosestToProperties
// 
// Finds the display mode in the master list closest to the one specified.  A semi-fuzzy
//	"scoring" system is used to evaluate each mode in the table (in terms of its similarity
//	to the properties described); the best-scoring mode "wins" and is returned.
// 
// Note that the scoring system uses the specified properties in the following order
//	of importance for finding a match (from most important to least):
// 
// - windowed vs. fullscreen
// - size (width, then height)
// - bit depth
// - display refresh frequency
// 
// For example, given a request for (800x600-32bpp-75hz), the mode (800x600-32bpp-60hz)
//	would score higher than the mode (800x600-16bpp-75hz) - both of whom would lose out to
//	an exact match, if present.
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::_GetVidModeClosestToProperties(
	const int width, const int height, const int bpp, const int frequency, const unsigned int flags )
{
	int bestMode = 0;
	int bestScore = 0;
	bool isWindowed = !(flags & VIDMODE_FLAG_FULLSCREEN);

	/// Determine the target frequency
	int targetFrequency = frequency;
	if( targetFrequency == 0 )
		targetFrequency = 60;

	/// Score each display mode in terms of its similarity to the properties specified
	for( unsigned int i = 0; i < _numVidModes; ++ i )
	{
		/// Assume a perfect score to start
		int score = 1000000000; // best possible score, for a perfect match
		VidModeDesc& mode = *_vidModes[ i ];

		/// Penalize for a windowed / fullscreen mismatch
		if( isWindowed != IsVidModeWindowed( i ) )
			score -= 500000000;

		/// Penalize for a width mismatch
		int widthDifference = abs( width - mode.devMode.dmPelsWidth );
		score -= widthDifference * 10000;

		/// Penalize for a height mismatch
		int heightDifference = abs( height - mode.devMode.dmPelsHeight );
		score -= heightDifference * 1000;

		/// Only do frequency & bpp matching if looking for a fullscreen mode
		if( !isWindowed )
		{
			/// Penalize for a bpp mismatch (but only for fullscreen modes)
			if( unsigned int(bpp) != mode.devMode.dmBitsPerPel )
				score -= 500;

			/// Penalize for a refresh frequency mismatch if one was specified
			int frequencyDifference = abs( targetFrequency - mode.devMode.dmDisplayFrequency );
			score -= frequencyDifference * 1;
		}

		/// Check if this score beats the best yet seen
		if( score > bestScore )
		{
			bestScore = score;
			bestMode = i;
		}
	}

	return bestMode;
}


//-------------------------------------------------------------------------------------------------
// _CreateWindowedDisplayMode
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::_CreateWindowedDisplayMode( const int width, const int height, const int bpp )
{
	/// Make sure the vidMode table isn't full
	VALIDATE_MSG1( _numVidModes < MAX_VID_MODES, "TheVideoSystem::_CreateWindowedDisplayMode : had to stop enumerating vid\nmodes; table was full (max %d entries!)\n", MAX_VID_MODES );

	/// Allocate a new entry and advance the count
	unsigned int newVidModeNum = _numVidModes;
	_vidModes[ newVidModeNum ] = new VidModeDesc;
	_numVidModes ++;

	/// Set the appropriate member fields to describe this mode
	VidModeDesc& vidModeDesc = *_vidModes[ newVidModeNum ];
	memset( &vidModeDesc.devMode, 0, sizeof( vidModeDesc.devMode ) );
	vidModeDesc.flags = 0;
	vidModeDesc.devMode.dmSize = sizeof( vidModeDesc.devMode );
	vidModeDesc.devMode.dmPelsWidth = width;
	vidModeDesc.devMode.dmPelsHeight = height;
	vidModeDesc.devMode.dmBitsPerPel = bpp;
	vidModeDesc.devMode.dmDisplayFrequency = 0; // will this cause problems later on?
	vidModeDesc.devMode.dmFields = DM_BITSPERPEL | DM_PELSWIDTH | DM_PELSHEIGHT; // says which fields we've filled in

	return newVidModeNum;
}


//-------------------------------------------------------------------------------------------------
// SetCurrentVidMode
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::SetCurrentVidMode( const unsigned int vidMode, const bool forceVidRestart )
{
	/// Check to make sure <vidMode> isn't out of range
	VALIDATE_MSG2( vidMode >= 0, SUBSYSTEM_NAME" : vidMode %d out of range [0,%d]!", vidMode, _numVidModes-1 );
	VALIDATE_MSG2( vidMode < _numVidModes, SUBSYSTEM_NAME" : vidMode %d out of range [0,%d]!", vidMode, _numVidModes-1 );

	/// Check if this mode is already the current one (unless <forceVidRestart> is true)
	if( !forceVidRestart )
	{
		if( vidMode == _currentVidMode )
			return;
	}

	/// Set this mode as the new current mode
	theConsole.Printf( CHAN_VIDEO | CHAN_HIGH, "VideoSystem : Setting current display mode #%d: %s\n", vidMode, GetVidModeDescription( vidMode ).c_str() );
	_currentVidMode = vidMode;

	/// Check if system is already running
	if( _isInitialized )
	{
		/// Destroy the rendering context and do a vid restart
		PANIC( "We're sorry, but the engine does not currently support run-time alteration of video mode.\n" );
	}
}


//-------------------------------------------------------------------------------------------------
// GetHWnd
//-------------------------------------------------------------------------------------------------
//const HWND TheVideoSystem::GetHWnd() const
STATIC const HWND TheVideoSystem::GetHWnd()
{
	return _hWnd;
}


//-------------------------------------------------------------------------------------------------
// QueryCurrentVidModeProperties
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::QueryCurrentVidModeProperties(
	int* pWidth, int* pHeight, int* pBpp, int* pFrequency, unsigned int* pFlags ) const
{
	/// Get some convenience references
	VidModeDesc& vidMode = *_vidModes[ _currentVidMode ];
	DEVMODE& devMode = vidMode.devMode;

	/// Get the current display mode width, if requested
	if( pWidth )
		*pWidth = devMode.dmPelsWidth;

	/// Get the current display mode height, if requested
	if( pHeight )
		*pHeight = devMode.dmPelsHeight;

	/// Get the current display mode bit depth, if requested
	if( pBpp )
		*pBpp = devMode.dmBitsPerPel;

	/// Get the current display mode frequency, if requested
	if( pFrequency )
		*pFrequency = devMode.dmDisplayFrequency;

	/// Get the current display mode flags, if requested
	if( pFlags )
		*pFlags = vidMode.flags;
}


//-------------------------------------------------------------------------------------------------
// GetCurrentVidModeWidth
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::GetCurrentVidModeWidth() const
{
	return( _vidModes[ _currentVidMode ]->devMode.dmPelsWidth );
}


//-------------------------------------------------------------------------------------------------
// GetCurrentVidModeHeight
//-------------------------------------------------------------------------------------------------
const int TheVideoSystem::GetCurrentVidModeHeight() const
{
	return( _vidModes[ _currentVidMode ]->devMode.dmPelsHeight );
}


//-------------------------------------------------------------------------------------------------
// GetVidModeDescription
//-------------------------------------------------------------------------------------------------
const string TheVideoSystem::GetVidModeDescription( int vidMode ) const
{
	DEVMODE& dm = _vidModes[ vidMode ]->devMode;
	/// FIXME: make a version of Sprintf that takes an STL string as a target
	char temp[ 256 ];
	Sprintf( temp, 256, "%d x %d x %d bpp", dm.dmPelsWidth, dm.dmPelsHeight, dm.dmBitsPerPel );
	string description( temp );

	if( IsVidModeWindowed( vidMode ) )
	{
		description += " (windowed)";
	}
	else
	{
		Sprintf( temp, 256, ", %d hz (fullscreen)", dm.dmDisplayFrequency );
		description += temp;
	}

	return description;
}


//-------------------------------------------------------------------------------------------------
// NotifyOfWindowMove
//-------------------------------------------------------------------------------------------------
void TheVideoSystem::NotifyOfWindowMove( int newClientX, int newClientY )
{
	_clientOffset.x = newClientX;
	_clientOffset.y = newClientY;
}


