// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.hpp"
#include "Baby.hpp"
#include "TheMap.hpp"
#include "Mine.hpp"


using namespace std;


//-------------------------------------------------------------------------------------------------
// Class constants
//-------------------------------------------------------------------------------------------------
STATIC const float Baby::BABY_CRAWL_SPEED				= 50.0f;
STATIC const float Baby::BABY_RADIUS					= 20.0f;
STATIC const char* Baby::BABY_ANIM_NAME_IDLE			= "Baby/Baby_idle.tga";
STATIC const char* Baby::BABY_ANIM_NAME_CRAWL			= "Baby/Baby_crawl.tga";
STATIC const char* Baby::BABY_ANIM_NAME_WOUNDEDIDLE		= "Baby/Baby_idle.tga";
STATIC const char* Baby::BABY_ANIM_NAME_WOUNDEDCRAWL	= "Baby/Baby_idle.tga";
STATIC const char* Baby::BABY_ANIM_NAME_DEAD			= "Baby/Baby_dead.tga";
STATIC const char* Baby::BABY_ANIM_NAME_HEART			= "Baby/RedBabyHeartIcon.tga" ;
STATIC const dword_t	Baby::BABY_THINK_DELAY			= 10000 ;
STATIC const unsigned int Baby::BABY_STARTING_HEARTS	= 2 ;

STATIC std::vector<Baby*> Baby::babyList;


//-------------------------------------------------------------------------------------------------
// Constructor
//-------------------------------------------------------------------------------------------------
Baby::Baby( const Vector2& startPosition, float startFacingDegrees )
:
//	anim_idle( animIdlePath ),
//	anim_woundedCrawl( animWoundedCrawlPath )
	position( startPosition ),
	_headingDegrees( startFacingDegrees ),
	_directionVector( 0.0f, 0.0f ),
	_lastThinkTime( 0 ),
	_speed( BABY_CRAWL_SPEED ),
	_hearts( BABY_STARTING_HEARTS ),
	_babyState( BABY_STATE_ALIVE )
{
	/// Register myself with the master list
	babyList.push_back( this );
	_babyNumber = babyList.size();

	anim_crawl	= new Animation( BABY_ANIM_NAME_CRAWL );
	anim_dead	= new Animation( BABY_ANIM_NAME_DEAD );
	anim_heart	= new Animation( BABY_ANIM_NAME_HEART );
}


//-------------------------------------------------------------------------------------------------
// Destructor
//-------------------------------------------------------------------------------------------------
Baby::~Baby()
{
	/// Remove myself from the master list
	std::vector<Baby*>::iterator iter;
	for( iter = babyList.begin(); iter != babyList.end(); iter ++ )
	{
		if( *iter == this )
		{
			babyList.erase( iter );
			break;
		}
	}
}


//-------------------------------------------------------------------------------------------------
// UpdateAll
//-------------------------------------------------------------------------------------------------
STATIC void Baby::UpdateAll()
{
	std::vector<Baby*>::iterator iter;
	for( iter = babyList.begin(); iter != babyList.end(); iter ++ )
	{
		Baby* baby = *iter;
		baby->Update();
	}
}


//-------------------------------------------------------------------------------------------------
// Update
//-------------------------------------------------------------------------------------------------
void Baby::Update()
{
	if ( _babyState == BABY_STATE_ALIVE )
	{
		_think();
		_move();
		_checkForCollisions();
	}
}



//-------------------------------------------------------------------------------------------------
// This imparts an immediate concussion/impact on the baby
//-------------------------------------------------------------------------------------------------
void Baby::concussionPoint( Vector2 &impactPoint, float radius )
{
	
}



//-------------------------------------------------------------------------------------------------
// DrawAll
//-------------------------------------------------------------------------------------------------
STATIC void Baby::DrawAll()
{
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	std::vector<Baby*>::iterator iter;
	for( iter = babyList.begin(); iter != babyList.end(); iter ++ )
	{
		Baby* baby = *iter;
		baby->Draw();
	}
}


//-------------------------------------------------------------------------------------------------
// Draw
//-------------------------------------------------------------------------------------------------
void Baby::Draw()
{
	float headingDegrees	= _headingDegrees - 90.0f ;

	if ( theMap.getTerrainAt( position ) == TERRAIN_ICE )
	{
		headingDegrees = RandomFloat( 360.0f );
	}

	if ( _babyState == BABY_STATE_ALIVE )
	{
		anim_crawl->Draw( position, Vector2( 0.5f, 0.5f), headingDegrees );
	}
	else
	{
		anim_dead->Draw( position, Vector2( 0.5f, 0.5f ), headingDegrees );
	}

//	theRenderer.DrawCircle( position, BABY_RADIUS, Rgba::GREEN );
//	theRenderer.DrawVector( position, _goalPosition );

	float yPosition	= 768.0f - ( (float)_babyNumber*32.0f);
	float xPosition = 32.0f ;
	Vector2 heartPosition( xPosition, yPosition );
	for ( unsigned int heart = _hearts; heart > 0; --heart )
	{
		anim_heart->Draw( heartPosition, Vector2( 0.5f, 0.5f ) );
		heartPosition.x += 64 ;
	}
}


STATIC bool Baby::BabiesAllDead( void )
{
	std::vector<Baby*>::iterator	iter ;
	for ( iter = babyList.begin(); iter != babyList.end(); ++iter )
	{
		Baby* baby = (*iter);
		if ( baby->_hearts > 0 )
			return false ;
	}
	return true ;
}


//-------------------------------------------------------------------------
//                P R O T E C T E D   M E T H O D S
//-------------------------------------------------------------------------

//------------------------------------------------------------------------
// If it is time, we pick a new goal
//------------------------------------------------------------------------
void Baby::_think( void )
{
	if ( _lastThinkTime < Timer::GetSysTimeMS() )
	{
		_pickNewGoal();
		_lastThinkTime = Timer::GetSysTimeMS() + BABY_THINK_DELAY ;
	}
}



//-------------------------------------------------------------------------
// Moves a baby
//-------------------------------------------------------------------------
void Baby::_move( void )
{
	// move towards goal
	if ( EpsilonEqual( position, _goalPosition, 1.0f ) )
	{
		_lastThinkTime = 0 ;
	}
	else 
	{
		position += _directionVector ;
	}
}


//--------------------------------------------------------------------------
// Baby picks a new goal
//--------------------------------------------------------------------------
void Baby::_pickNewGoal( void )
{
	bool	lookingForGoal	= true ;
	float	xOffset			= 0.0f ;
	float	yOffset			= 0.0f ;
	float	headingDegrees	= 0.0f ;

	while ( lookingForGoal )
	{
		headingDegrees		= RandomFloat( 360.0f );
		xOffset				= 200.0f * (float)cos( DEG2RAD( headingDegrees ) ) + position.x ;
		yOffset				= 200.0f * (float)sin( DEG2RAD( headingDegrees ) ) + position.y ;

		if ( xOffset > 0 && xOffset < TheMap::MAP_WIDTH && yOffset > 0 && yOffset < TheMap::MAP_HEIGHT )
		{
			Vector2 tempGoalPosition( xOffset, yOffset );
			if ( theMap.tracePath( position, tempGoalPosition, BABY_RADIUS ) )
			{
				lookingForGoal = false ;
			}
		}
	}

	_goalPosition.SetXY( xOffset, yOffset );
	_headingDegrees = headingDegrees ;

	_directionVector				= _goalPosition - position ;
	float fs = (float) theApplication.GetFrameSeconds();
	if ( fs > 0.1f )
	{
		fs = 0.1f ;
	}
	float moveDistance = _speed * fs ;
	_directionVector.SetLength( moveDistance );
}



void Baby::_checkForCollisions( void )
{
	Mine* mine		= Mine::FindFirstMineThatCollidesWithPoint( position, BABY_RADIUS );

	if ( mine )
	{
		_hearts-- ;
		_speed *= 0.85f ;
		delete mine ;
	}

	if ( _hearts <= 0 )
	{
		_babyState = BABY_STATE_DEAD ;
	}
}

