// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.

#include "pch.hpp"
#include "gameCommon.hpp"
#include "Firework.hpp"
using namespace std;


//-------------------------------------------------------------------------------------------------
// Class constants and static variable initialization
//-------------------------------------------------------------------------------------------------
#define FIREWORK_ENTRY_KEYWORD "firework"
#define FIREWORKS_DIRECTORY_NAME "Fireworks/"
STATIC const char* Firework::SEARCH_FILE_PATTERN = FIREWORKS_DIRECTORY_NAME"FireworkData_*.txt";
STATIC std::vector<Firework> Firework::m_fireworkDefinitions;


KeyTypeTableEntry keyTypeTable[] =
{
//	keyName				keyType
	"fuseSeconds",		KEYTYPE_FLOAT_RANGE,
	"durationSeconds",	KEYTYPE_FLOAT_RANGE,
	"attractValue",		KEYTYPE_FLOAT_RANGE,
	"attractRadius",	KEYTYPE_FLOAT_RANGE,
	"blastForce",		KEYTYPE_FLOAT_RANGE,
	"blastDamage",		KEYTYPE_FLOAT_RANGE,
	"blastRadius",		KEYTYPE_FLOAT_RANGE,
	"iconImage",		KEYTYPE_STRING,
	"projectileImage",	KEYTYPE_STRING,
};
const int keyTypeTableSize = sizeof( keyTypeTable ) / sizeof( keyTypeTable[0] );



//-------------------------------------------------------------------------------------------------
// IsANumberChar
//-------------------------------------------------------------------------------------------------
bool IsANumberChar( char test )
{
	if( isdigit( test ) )
		return true;

	if( test == '.' )
		return true;

	if( test == '-' )
		return true;

	return false;
}


//-------------------------------------------------------------------------------------------------
// constructor
//-------------------------------------------------------------------------------------------------
Firework::Firework()
:
fuseSecondsMinMax( Vector2::ZERO ),
durationSecondsMinMax( Vector2::ZERO ),
attractValueMinMax( Vector2::ZERO ),
attractRadiusMinMax( Vector2::ZERO ),
blastForceMinMax( Vector2::ZERO ),
blastDamageMinMax( Vector2::ZERO ),
blastRadiusMinMax( Vector2::ZERO )
{
}


//-------------------------------------------------------------------------------------------------
// destructor
//-------------------------------------------------------------------------------------------------
Firework::~Firework()
{
}


//-------------------------------------------------------------------------------------------------
// LoadFireworkFile
// 
// Returns the number of entries parsed from the file.
//-------------------------------------------------------------------------------------------------
STATIC int Firework::ParseFireworkFile( const FileData& fileData )
{
	int entryCount = 0;

	/// Get the full path
	string pathName( string( FIREWORKS_DIRECTORY_NAME ) + fileData.name );

	/// Open the file
	FILE* file = fopen( pathName.c_str(), "rb" );
	if( !file )
	{
		PANIC_MSG1( "Firework::ParseFireworkFile : ERROR : couldn't open file %s\n", pathName.c_str() );
		return 0;
	}

	/// Get the file size
	fseek( file, 0, SEEK_END );
	int fileSize = ftell( file );
	rewind( file );

	/// Read the file into a text buffer
	char* buffer = new char[ fileSize + 1 ];
	fread( buffer, 1, fileSize, file );
	buffer[ fileSize ] = '\0';

	/// Close the file
	fclose( file );

	/// Parse the buffer
	entryCount += ParseFireworkBuffer( buffer );

	/// Free the buffer
	delete[] buffer;
	buffer = NULL;

	return entryCount;
}


//-------------------------------------------------------------------------------------------------
// ParseAllFireworkFiles
//-------------------------------------------------------------------------------------------------
STATIC void Firework::ParseAllFireworkFiles()
{
	int fileCount = 0;
	int entryCount = 0;
	long findHandle;
	FileData fileData;

	/// Find all files in the /fireworks directory matching the search pattern
	findHandle = _findfirst( SEARCH_FILE_PATTERN, &fileData );
	if( findHandle == -1 )
	{
		return; // no firework data files found!
	}

	/// Load all files matching the search string
	for( ;; )
	{
		entryCount += ParseFireworkFile( fileData );
		fileCount ++;

		int result = _findnext( findHandle, &fileData );
		if( result == -1 )
		{
			break;
		}
	}

	/// End the search
	_findclose( findHandle );

	theConsole.Printf( "ParseAllFireworkFiles : parsed %d entries in %d files.\n", entryCount, fileCount );
}


//-------------------------------------------------------------------------------------------------
// ParseFireworkBuffer
//-------------------------------------------------------------------------------------------------
STATIC int Firework::ParseFireworkBuffer( const char* buffer )
{
	ASSERT_MSG0( buffer, "Firework::ParseFireworkBuffer : ERROR : buffer was NULL!\n" );

	int entryCount = 0;

	const char* read = buffer;
	for( ;; )
	{
		/// Look for firework keyword
		char* scan = strstr( read, FIREWORK_ENTRY_KEYWORD );
		if( !scan )
		{
			break; // no more entries to scan
		}

		/// Parse firework definition block
		Firework tempObject;
		read = tempObject.ParseFireworkDefinition( read );

		/// Add a copy of this temp object to the static vector list of firework types
		m_fireworkDefinitions.push_back( tempObject );

		entryCount ++;
	}

	return entryCount;
}


//-------------------------------------------------------------------------------------------------
// ParseFireworkDefinition
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseFireworkDefinition( const char* read )
{
	int keywordLen = strlen( FIREWORK_ENTRY_KEYWORD );

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Skip keyword
	read += keywordLen;

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Scan firework name until non-whitespace
	char fireworkName[ 1024 ];
	char* write = fireworkName;
	while( !isspace( *read ) )
	{
		*write = *read;
		write ++;
		read ++;
	}
	*write = '\0';

	/// Set the firework's name
	m_fireworkName = fireworkName;

	/// Scan until opening brace
	read = strchr( read, '{' );
	ASSERT_MSG1( read, "ParseFireworkDefinitionBlock: PARSE ERROR : couldn't find opening brace for entry %s\n", m_fireworkName );

	read = ParseFireworkDefinitionBlock( read );
	return read;
}


//-------------------------------------------------------------------------------------------------
// ParseFireworkDefinitionBlock
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseFireworkDefinitionBlock( const char* read )
{
	/// Scan past brace
	read ++;

	for( ;; )
	{
		/// Scan past whitespace to next entry (or closing brace)
		while( isspace( *read ) ) read ++;

		ASSERT_MSG1( *read, "ParseFireworkDefinitionBlock: PARSE ERROR : entry %s had no closing brace (unexpected end-of-file)!\n", m_fireworkName );

		/// Check for closing brace
		if( *read == '}' )
		{
			/// Advance past closing brace and return
			read ++;
			break;
		}

		/// Must be a key/value line; parse it
		read = ParseKeyValueLine( read );
	}

	return read;
}


//-------------------------------------------------------------------------------------------------
// ParseKeyValueLine
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseKeyValueLine( const char* read )
{
	char keyName[ 1024 ];
	char* write = keyName;

	/// Grab the key name
	while( !isspace( *read ) && ( *read != '=' ) )
	{
		*write = *read;
		write ++;
		read ++;
	}
	*write = '\0';

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Skip the '='
	ASSERT_MSG2( *read == '=', "ParseFireworkDefinitionBlock: PARSE ERROR : key %s in entry %s was missing an '=' sign!\n", keyName, m_fireworkName );
	read ++;

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Parse the value (may be a number, a number range, a string, etc)
	read = ParseValueForKey( read, keyName );

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	return read;
}


//-------------------------------------------------------------------------------------------------
// GetKeyTypeForKeyName
//-------------------------------------------------------------------------------------------------
KeyType Firework::GetKeyTypeForKeyName( const char* keyName )
{
	for( int i = 0; i < keyTypeTableSize; i ++ )
	{
		if( !stricmp( keyName, keyTypeTable[ i ].keyName ) )
		{
			return keyTypeTable[ i ].keyType;
		}
	}

	return KEYTYPE_INVALID;
}



//-------------------------------------------------------------------------------------------------
// ParseValueForKey
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseValueForKey( const char* read, const char* keyName )
{
	/// Figure out what key type this is
	KeyType keyType = GetKeyTypeForKeyName( keyName );

	switch( keyType )
	{
		case KEYTYPE_FLOAT_RANGE:
			read = ParseValueForKey_FloatRange( read, keyName );
			break;

		case KEYTYPE_STRING:
			read = ParseValueForKey_String( read, keyName );
			break;

		default:
			PANIC_MSG2( "Firework::ParseValueForKey : PARSE ERROR : in entry \"%s\", didn't recognize data key named \"%s\"!\n", m_fireworkName.c_str(), keyName );
			break;
	}

	return read;
}


//-------------------------------------------------------------------------------------------------
// ParseFloatValue
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseFloatValue( const char* read, float& value )
{
	char valueString[ 1024 ];
	char* write = valueString;

	while( IsANumberChar( *read ) )
	{
		*write = *read;
		write ++;
		read ++;
	}
	*write = '\0';

	value = (float) atof( valueString );
	return read;
}


//-------------------------------------------------------------------------------------------------
// ParseFloatRange
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseFloatRange( const char* read, float& min, float& max )
{
	/// First character should be '['; skip it
	ASSERT_MSG1( *read == '[', "Firework::ParseValueForKey_String : in entry %s, couldn't find expected '['\n", m_fireworkName.c_str() );
	read ++;

	/// Read the first float
	read = ParseFloatValue( read, min );

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Skip the expected comma
	ASSERT_MSG1( *read == ',', "Firework::ParseValueForKey_String : in entry %s, couldn't find expected ','\n", m_fireworkName.c_str() );
	read ++;

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Read the first float
	read = ParseFloatValue( read, max );

	/// Skip whitespace
	while( isspace( *read ) ) read ++;

	/// Skip the expected closing ']'
	ASSERT_MSG1( *read == ']', "Firework::ParseValueForKey_String : in entry %s, couldn't find expected ']'\n", m_fireworkName.c_str() );
	read ++;

	return read;
}



//-------------------------------------------------------------------------------------------------
// ParseValueForKey_FloatRange
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseValueForKey_FloatRange( const char* read, const char* keyName )
{
	float min = 0.0f;
	float max = 0.0f;

	/// The next character should be either a number or '[' (for a number range)
	if( *read == '[' )
	{
		read = ParseFloatRange( read, min, max );
	}
	else if( IsANumberChar( *read ) )
	{
		float value;
		read = ParseFloatValue( read, value );
		min = max = value;
	}
	else
	{
		PANIC_MSG1( "Firework::ParseValueForKey_FloatRange : found unexpected character '%c'!\n", *read );
		return read;
	}

	/// Set the appropriate member
	SetDataMember( keyName, min, max );

	return read;
}


//-------------------------------------------------------------------------------------------------
// ParseValueForKey_String
//-------------------------------------------------------------------------------------------------
const char* Firework::ParseValueForKey_String( const char* read, const char* keyName )
{
	/// The next character should be a quotation mark
	if( *read != '\"' )
	{
		PANIC_MSG2( "Firework::ParseValueForKey_String : in entry %s, key %s, couldn't find quote after = !\n", m_fireworkName.c_str(), keyName );
		return read;
	}

	/// Skip it and scan until quotation mark
	read ++;
	char valueString[ 1024 ];
	char* write = valueString;
	while( *read && *read != '\"' )
	{
		*write = *read;
		write ++;
		read ++;
	}
	*write = '\0';

	/// Skip closing quotation
	read ++;

	SetDataMember( keyName, valueString );

	return read;
}


//-------------------------------------------------------------------------------------------------
// SetDataMember
//-------------------------------------------------------------------------------------------------
void Firework::SetDataMember( const char* keyName, float min, float max )
{
	if( !stricmp( keyName, "fuseSeconds" ) )
	{
		fuseSecondsMinMax.SetXY( min, max );
	}
	else if( !stricmp( keyName, "durationSeconds" ) )
	{
		durationSecondsMinMax.SetXY( min, max );
	}
	else if( !stricmp( keyName, "attractValue" ) )
	{
		attractValueMinMax.SetXY( min, max );
	}
	else if( !stricmp( keyName, "attractRadius" ) )
	{
		attractRadiusMinMax.SetXY( min, max );
	}
	else if( !stricmp( keyName, "blastForce" ) )
	{
		blastForceMinMax.SetXY( min, max );
	}
	else if( !stricmp( keyName, "blastDamage" ) )
	{
		blastDamageMinMax.SetXY( min, max );
	}
	else if( !stricmp( keyName, "blastRadius" ) )
	{
		blastRadiusMinMax.SetXY( min, max );
	}
	else
	{
		PANIC_MSG2( "Firework::SetDataMember : PARSE ERROR : in entry %s, key %s was NOT a float min/max!\n", m_fireworkName.c_str(), keyName );
	}
}


//-------------------------------------------------------------------------------------------------
// SetDataMember
//-------------------------------------------------------------------------------------------------
void Firework::SetDataMember( const char* keyName, const char* stringValue )
{
	if( !stricmp( keyName, "iconImage" ) )
	{
		iconImageString = stringValue;
	}
	else if( !stricmp( keyName, "projectileImage" ) )
	{
		projectileImageString = stringValue;
	}
	else
	{
		PANIC_MSG2( "Firework::SetDataMember : PARSE ERROR : in entry %s, key %s was NOT a string value!\n", m_fireworkName.c_str(), keyName );
	}
}



