//----------------------------------------------------
// TheMap.cpp -- Implementation of TheMap class.
// 
// Copyright 2004 ALGDS, Zero Lodge.  See Engine/License.hpp for usage and license details.
//----------------------------------------------------


#include "pch.hpp"
#include "TheMap.hpp"
#include "Baby.hpp"
#include "Mine.hpp"
#include "Tower.hpp"
#include "gamecommon.hpp"

// statics and const
const unsigned int TheMap::MAP_HEIGHT		= 768 ;
const unsigned int TheMap::MAP_WIDTH		= 1024 ;
const unsigned int TheMap::BABY_START		= 0x0000ff00 ;	// GREEN
const unsigned int TheMap::TOWER_SPAWN		= 0x00ffff00 ;	// YELLOW
const unsigned int TheMap::MINE_SPAWN		= 0x00ff0000 ;	// RED
const unsigned int TheMap::GRINDER_SPAWN	= 0x00ff00ff ;	// MAGENTA
const unsigned int TheMap::OBSTACLE			= 0x00ffffff ;	// WHITE
const unsigned int TheMap::ICE				= 0x0000ffff ;	// CYAN
const unsigned int TheMap::MUD				= 0x00969600 ;	// BROWN

// The extern'ed singleton
TheMap theMap ;


//---------------------------------
// Constructor
//---------------------------------
TheMap::TheMap( void )
	:	_mapData( NULL ),
		_mapBackground( NULL ),
		_mapImageData( NULL ),
		_isInitialized( false )
{

}


//---------------------------------
// Destructor
//---------------------------------
TheMap::~TheMap( void )
{
	if ( _mapData )
	{
		delete _mapData ;
		_mapData = NULL ;
	}
}


//-----------------------------------------
// Returns true if the specified map position is blocked
//-----------------------------------------
bool TheMap::isBlocked( Vector2& position )
{
	return isBlocked( (unsigned int)position.x, (unsigned int)position.y );
}

//-----------------------------------------
// Returns true if the specified map position is blocked
//-----------------------------------------
bool TheMap::isBlocked( unsigned int x, unsigned int y )
{
	ASSERT_MSG0( _mapData != NULL, "Map Data uninitialized\n" );
	ASSERT_MSG1( x < MAP_WIDTH, "X coordinate out of range: %d\n", x );
	ASSERT_MSG1( y < MAP_HEIGHT, "Y coordinate out of range: %d\n", y );

	return ( _mapData[ x + (y * MAP_WIDTH ) ]._terrainType == TERRAIN_BLOCKED ? true : false );
}


//--------------------------------------------
// Retrieves the terrain type at the specified vector position
//--------------------------------------------
TerrainType TheMap::getTerrainAt( Vector2& position )
{
	return getTerrainAt( (unsigned int)position.x, (unsigned int)position.y );
}

//--------------------------------------------
// Retrieves the terrain type at the specified position
//--------------------------------------------
TerrainType TheMap::getTerrainAt( unsigned int x, unsigned int y )
{
	ASSERT_MSG0( _mapData != NULL, "Map Data uninitialized\n" );
	ASSERT_MSG1( x < MAP_WIDTH, "X coordinate out of range: %d\n", x );
	ASSERT_MSG1( y < MAP_HEIGHT, "Y coordinate out of range: %d\n", y );

	return _mapData[ x + (y * MAP_WIDTH ) ]._terrainType ;
}



//----------------------------------------
// Loads a level from the specified filename.
//----------------------------------------
void TheMap::loadLevel( const std::string& name )
{
	_name = name ;
	_init();

	// get level data
	_mapBackground		= new Animation( LEVEL_DIRECTORY + name + "_image.tga" );
	_mapImageData		= new Image( LEVEL_DIRECTORY + name + "_data.tga" );

	_parseMapData();
}


//--------------------------------------------
// Updates the map.  Currently does nothing.
//--------------------------------------------
bool TheMap::update( void )
{
	return true ;
}



//--------------------------------------------
// Draws the background
//--------------------------------------------
void TheMap::draw( void )
{
	if ( _mapBackground )
	{
		Vector2 texCoordTopLeft( 0.0f, 0.25f );
		Vector2 texCoordBottomRight( 1.0f, 1.0f );
		//theRenderer.DrawTexturedRect( *_mapBackground, Vector2(0.0f, 0.0f), Vector2( 1024.0f, 768.0f ), texCoordTopLeft, texCoordBottomRight );
		_mapBackground->Draw(Vector2(512.0f, 256.0f));
	}

	return ;
	glPointSize( 3.0f );
	glBlendFunc( GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA );
	glEnable( GL_BLEND );

	glBegin( GL_POINTS );
	glColor3f( 1.0f, 0.5f, 0.5f );
	for ( unsigned int y = 0; y < MAP_HEIGHT; ++y )
		for ( unsigned int x = 0; x < MAP_WIDTH; ++x )
		{
			if ( _mapData[ x + (y * MAP_WIDTH )]._terrainType == TERRAIN_BLOCKED )
			{
				glVertex2f( (float)x, (float)y );
			}
		}
	glEnd();


}

bool TheMap::tracePath( Vector2& startPosition, Vector2& endPosition, float radius )
{
	Vector2 directionVector ;
	Vector2 tempPosition( startPosition );

	directionVector = endPosition - startPosition ;
	directionVector.SetLength( radius );

	while ( !EpsilonEqual( tempPosition, endPosition, radius ) )
	{
		tempPosition	+= directionVector ;
		if ( !_isPositionValid( (unsigned int)tempPosition.x, (unsigned int)tempPosition.y, radius ) )
			return false ;
	}

	return true ;
}





//--------------------------------------------------------------
//         P R O T E C T E D   M E T H O D S
//--------------------------------------------------------------

//----------------------------------
// Initializes the map.  Allocates data.
//----------------------------------
void TheMap::_init( void )
{
	_mapData	= new MapCell[ 1024 * 768 ] ;
	for ( unsigned int y=0; y < MAP_HEIGHT; ++y )
	{
		for ( unsigned int x=0; x < MAP_WIDTH; ++x )
		{
			_mapData[ x + (y * MAP_WIDTH ) ]._terrainType =  TERRAIN_CLEAR ;
		}
	}
}


//--------------------------------------------------------------
// Parses the map data.
//--------------------------------------------------------------
void TheMap::_parseMapData( void )
{
	if ( !_mapImageData )		return ;

	unsigned char*	pixelData	= _mapImageData->getPixelData() ;
	unsigned char	pixelB		= 0;
	unsigned char	pixelG		= 0;
	unsigned char	pixelR		= 0;
	unsigned int	pixel		= 0;
	unsigned int	width		= _mapImageData->getWidth() ;
	unsigned int	height		= _mapImageData->getHeight() ;
	TerrainType		terrainType	= TERRAIN_CLEAR ;

	for ( unsigned int y = 0; y < MAP_HEIGHT; ++y )
	{
		for ( unsigned int x = 0; x < MAP_WIDTH; ++x )
		{
			unsigned imagePos	= ((y + 256)* width + x ) * 3 ;
			pixelB		= pixelData[ imagePos + 0 ] ;
			pixelG		= pixelData[ imagePos + 1 ];
			pixelR		= pixelData[ imagePos + 2 ];

			pixel		= (0 << 24) + (pixelR << 16) + (pixelG << 8) + (pixelB << 0);
			switch ( pixel )
			{
				case BABY_START:
					terrainType = TERRAIN_CLEAR ;
					_spawnBaby( x, y );
					break ;
				case MINE_SPAWN:
					terrainType = TERRAIN_CLEAR ;
					_spawnMine( x, y );
					break ;
				case TOWER_SPAWN:
					terrainType	= TERRAIN_BLOCKED ;
					_spawnTower( x, y );
					break ;
				case GRINDER_SPAWN:
					terrainType = TERRAIN_BLOCKED ;
					break ;
				case OBSTACLE:
					terrainType = TERRAIN_BLOCKED ;
					break ;
				case ICE:
					terrainType = TERRAIN_ICE ;
					break ;
				case MUD:
					terrainType	= TERRAIN_MUD ;
					break ;
				default:
					terrainType	= TERRAIN_CLEAR ;
					break ;
			}
			_mapData[ x + (y * MAP_WIDTH ) ]._terrainType = terrainType ;
		}
	}
}



//-------------------------------------------------------
// Spawns a baby
//-------------------------------------------------------
void TheMap::_spawnBaby( unsigned int x, unsigned int y )
{
	Baby* baby = new Baby( Vector2( (float)x, (float)y), 0 );

	char message[ 256 ];
	sprintf( message, "Spawned baby at (%d,%d)\n", x, y );
	OutputDebugString( message );

	// babies register themselves so no need to track it.
}



//-------------------------------------------------------
// Spawns a map
//-------------------------------------------------------
void TheMap::_spawnMine( unsigned int x, unsigned int y )
{
	Mine* mine = new Mine( Vector2( (float)x, (float)y) );

	char message[ 256 ];
	sprintf( message, "Spawned mine at (%d,%d)\n", x, y );
	OutputDebugString( message );

	// mines register themselves so no need to track it.
}


//-----------------------------------------------------
// Spawns a new tower on the map.
//-----------------------------------------------------
void TheMap::_spawnTower( unsigned int x, unsigned int y )
{
	Tower* tower = new Tower( Vector2( (float)x, (float)y ) );

	char message[ 256 ];
	sprintf( message, "Spawned mine at (%d,%d)", x, y );
	OutputDebugString( message );
}


//--------------------------------------------------------------------------
// Returns true if the spot is accessible
//--------------------------------------------------------------------------
bool TheMap::_isPositionValid( unsigned int goalX, unsigned int goalY, float radius )
{
	int	startY	= goalY - (int)radius ;
	int endY	= goalY + (int)radius ;
	int startX	= goalX - (int)radius ;
	int endX	= goalX + (int)radius ;

	for ( int y = startY; y < endY; ++y )
	{
		for ( int x = startX; x < endX; ++x )
		{
			if ( x > 0 && x < (int)theMap.MAP_WIDTH && y > 0 && y < (int)theMap.MAP_HEIGHT )
			{
				if ( theMap.isBlocked( (unsigned int)x, (unsigned int)y ) )
				{
					return false ;
				}
			}
			else
			{
				return false ;
			}
		}
	}

	return true ;
}







